/** (c) Brendan Macmillan 2001, licensed under GNU's GPL 2 
    (Also available for non-GPL, commercial use, for fee) 
    Website: http://www.csse.monash.edu.au/~bren/JSX
    Website: http://freshmeat.net/projects/jsx
    List (can read without subscribing, need to join to post):
      http://groups.yahoo.com/group/JSX-ideas/messages
		Commercial licensing enquiries only:
			bren@mail.csse.monash.edu.au */


/** ===========================================================================
  * MagicClass  (JSX0.8.14.2)
	* ==========
	* Invoking the private native method allocationNewObject().
	* Also reading private etc fields.
	* NOTE: Class is final, so can't subclass it; make this MagicClass instead
	**/

package JSX;
import java.lang.reflect.*;	//for Method etc
import java.io.*;	//for Serializable interface

public class MagicClass {
		private final static boolean DEBUG = false;
		//private final static boolean DEBUG = true;

	/**------------------------------------------------------------------------
		* nativeCallInit
		* --------------
		* Factor out setup of all native calls needed.
		* Will need to cater for different JVM's here; for example, JDK1.4 goes
		* via a sun.reflect.ReflectionFactory to do allocateNewObject.
		* we need to detect failures, and try the next one.  It could be nice
		* to configure all this data in an XML file...
		*
		* Graceful degrade is simply to return null, if can't do it, and test
		* for this before use.
		**/
	static Method nativeCallInit(Class clazz, String methodName, Class[] argTypes) throws Exception {
		Method method = null;
		try {
			method = clazz.getDeclaredMethod(methodName, argTypes);
			method.setAccessible(true);
		} catch (NoSuchMethodException e) {
			if (DEBUG) {
				System.err.println("JSX couldn't find:");
				System.err.print("  "+clazz.getName()+"."+methodName+"(");
				for (int i=0; i<argTypes.length; i++) {
					System.err.print((i==0?"":", ")+argTypes[i].getName());
				}
				System.err.println(")");
				e.printStackTrace();
			}
			//degrade gracefully
		} catch (Exception e) {
			if (DEBUG) {
				System.err.println("Method: '"+methodName+"'");
				System.err.println("Arguments: ");
				for (int i=0; i<argTypes.length; i++)
					System.err.println("  "+argTypes[i]);
				System.err.print("Class: '"+clazz+"'");
				e.printStackTrace();
				System.exit(0);
			}
			throw e;
		}	

		return method;
	}


	/**------------------------------------------------------------------------
		* newInstance
		* -----------
		* Consider: subclass Class.  Cannot: "Class" is final
		**/
	/* We reuse the variable names from ObjectInputStream.inputObject(),
	 * purely for ease of reference with that method.
	 */
	private static Method allocateNewObject;
	private static Exception allocateNewObjectException;
	static {
		try {
			allocateNewObject = nativeCallInit(
				java.io.ObjectInputStream.class,		//class
				"allocateNewObject", new Class[] {Class.class, Class.class});	//method
			//now, we try the fall back position, for JDK1.4
		} catch (Exception e) {
			allocateNewObjectException = e;	//cork for later - based on Philip Gust
			//to adapt for JDK1.4 (and JDK1.2 if necessary)....
			if (allocateNewObject==null) {	//Exceptions interact badly with finals!
				//insert code here!
				allocateNewObject = null;
			}
		}
	}
	public static Object newInstance(Class currentClass)
			throws InvocationTargetException,
			       IllegalAccessException {
		if (allocateNewObject==null) {
			throw new InvocationTargetException(allocateNewObjectException);
			//NB: don't reset, because a static problem: fail once, fails *always*
		}
		//currentClass: is the actual runtime class we want to create.
		//currclass: is the closest non-serializable ancestor.
		// NB: Object is non-serializable, always bottoms out.
	//ONLY WORKS IF YOU ARE USING SERIALIZABLE CLASSES, I THINK!!!!
/*
		Class currclass = currentClass; 	//begin search for ancestor
		if (currclass != null) {
			while (Serializable.class.isAssignableFrom(currclass)) 
				currclass = currclass.getSuperclass();
		}		    
*/
		Class currclass = Object.class;	//always object
			//odd: why not use instanceOf?  <--- what did I mean by this?
 		Object[] argValues = {currentClass, currclass};
			//Downside: slower than reusing same array
		return allocateNewObject.invoke(null, argValues);
	}
//static private native Object allocateNewObject(Class currentClass, Class currclass);

	/**------------------------------------------------------------------------
		* setObjectFieldValue
		* -------------------
		**/
	private static Method setObjectFieldValue;	//final: must be set
	static {
		try {
			setObjectFieldValue = nativeCallInit(java.io.ObjectInputStream.class,
				"setObjectFieldValue",
				new Class[] {Object.class, long.class, Class.class, Object.class});
		} catch (Exception e) {
			//TODO: corking code.  See allocateNewObjectException
		}
	}
	public static void setObjectFieldValue(Object parent, ObjectStreamField f, Class type, Object value) {
		if (setObjectFieldValue==null) {
			return;	//graceful degrade
			//throw something instead?  Or *do* we want a graceful degrade?
		}
		try {
			ObjectStreamField[] fs = {f};	//know element type because of field
			if (DEBUG) {
				System.err.println("parent = " + parent);
				System.err.println("f = " + f);
				System.err.println("type = " + type);
				System.err.println("value = " + value);
			}
			long[] primIDs = new long[0];	//we know only one field, and it's an Object
			long[] objIDs = new long[1];
			getFieldIDs(fs, primIDs, objIDs);
			//System.err.println("primIDs = " + primIDs[0]);
			if (DEBUG) System.err.println("objIDs = " + objIDs[0]);
			Object[] argValues = {parent, new Long(objIDs[0]), type, value};
			setObjectFieldValue.invoke(null, argValues);
		} catch (Exception e) {
			System.err.println("Within JSX.MagicClass.setObjectFieldValue");
			e.printStackTrace();
				//a better policy may be to rethrow the wrapper exception...
		}
	}


	/**------------------------------------------------------------------------
		* setPrimitiveFieldValue
		* ----------------------
		**/
	private static Method setPrimitiveFieldValues;
	static {
		try {
			setPrimitiveFieldValues = nativeCallInit(java.io.ObjectInputStream.class,
				"setPrimitiveFieldValues",
				new Class[] {Object.class, long[].class, char[].class, byte[].class});
		} catch (Exception e) {
			//TODO: corking code.  See allocateNewObjectException
		}
	}
	public static void setPrimitiveFieldValues(Object parent, ObjectStreamField f, Object value) {
		if (setPrimitiveFieldValues==null)
			return;
		try {
			ObjectStreamField[] fs = {f};	//know element type because of field
			if (DEBUG) {
				System.err.println("parent = " + parent);
				System.err.println("f = " + f);
				//System.err.println("type = " + type);
				System.err.println("value = " + value);
			}
			long[] primIDs = new long[1];	//we know only one field, and it's an Object
			long[] objIDs = new long[1];
			getFieldIDs(fs, primIDs, objIDs);
			if (DEBUG) System.err.println("primIDs = " + primIDs[0]);
			//if (DEBUG) System.err.println("objIDs = " + objIDs[0]);
//This is *really* inefficient - but I can't see another way around it right
//now...
//(2). use appropriate "write*()" for that type.
//(3). Could - switch on all here, using a wrapper - *yes*, a wrapper used
//already, anyway.
//(4). Could - check how Java's own serialization does it.

			char[] typecode = {f.getTypeCode()};	//It already knows this
			ByteArrayOutputStream ba1 = new ByteArrayOutputStream();
			DataOutputStream ba = new DataOutputStream(ba1);
			switch (typecode[0]) {
				case 'B': ba.writeByte(((Byte)value).byteValue()); break; //byte
				case 'C': ba.writeChar(((Character)value).charValue()); break; //char
				case 'S': ba.writeShort(((Short)value).shortValue()); break; //short
				case 'I': ba.writeInt(((Integer)value).intValue()); break; //int
				case 'J': ba.writeLong(((Long)value).longValue()); break; //long
				case 'F': ba.writeFloat(((Float)value).floatValue()); break; //float
				case 'D': ba.writeDouble(((Double)value).doubleValue()); break; //double
				case 'Z': ba.writeBoolean(((Boolean)value).booleanValue()); break; //boolean
				case '[':	//an array...
				case 'L': //an object
					throw new Exception("setPrimitiveFieldValues only does primitives");
			}
			ba.flush();
			ba.close();
			Object[] argValues = {parent, primIDs, typecode, ba1.toByteArray()};
//			Object[] argValues = {parent, primIDs, typecode,
//				new byte[] {0x00,0x00,0x00,0x01}};
			setPrimitiveFieldValues.invoke(null, argValues);
		} catch (Exception e) {
			System.err.println("Within JSX.MagicClass.setObjectFieldValue");
			e.printStackTrace();
		}
	}

	/**------------------------------------------------------------------------
		* getObjectFieldValue
		* -------------------
		**/
	private static Method getObjectFieldValue;
	static {
		try {
			getObjectFieldValue = nativeCallInit(java.io.ObjectOutputStream.class,
				"getObjectFieldValue", new Class[] {Object.class, long.class});
		} catch (Exception e) {
			//TODO: corking code.  See allocateNewObjectException
		}
	}
/*
	public static Object getObjectFieldValue(Object parent, ObjectStreamField f) {
		try {
			ObjectStreamField[] fs = {f};	//know element type because of field
			long[] primIDs = new long[1];	//we know only one field, and it's an Object
			long[] objIDs = new long[1];
			getFieldIDs(fs, primIDs, objIDs);
			Object[] argValues = {parent, new Long(objIDs[0])};
			return getObjectFieldValue.invoke(null, argValues);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null; //should never get here
	}
*/

	/**------------------------------------------------------------------------
		* getFieldIDs
		* -----------
		**/
	private static Method getFieldIDs;
	static {
		try {
			getFieldIDs = nativeCallInit(java.io.ObjectStreamClass.class,
				"getFieldIDs",
				new Class[] {ObjectStreamField[].class, long[].class, long[].class});
		} catch (Exception e) {
			//TODO: corking code.  See allocateNewObjectException
		}
	}
	public static void getFieldIDs(ObjectStreamField[] fs, long[] primIDs, long[] objIDs) throws Exception, InvocationTargetException {
		if (getFieldIDs==null)
			return;
		//fs = null;  //exploration
		//primIDs = null;  //exploration
		//objIDs = null;  //exploration
		Object[] argValues = {fs, primIDs, objIDs};
		getFieldIDs.invoke(null, argValues); //null because static
	}

	/**------------------------------------------------------------------------
		* setIntField (com.sun.corba.se.internal.io.IIOPInputStream
		* -----------
		**/
/*  The following com.sun.* classes are not available under the GCC 3.0 gjc -
since they are not used, but left over from an experiment, I'm just going to
comment them out.  Many thanks to Tom Holroyd for pointing this out!*/
/*
	private static final Method setIntField;
	static {
		setIntField = nativeCallInit(com.sun.corba.se.internal.io.IIOPInputStream.class,
			"setIntField",
			new Class[] {Object.class, Class.class, String.class, String.class, int.class});
	}
	public static void setIntField(Object obj, Class clazz, String fieldName,
String fieldSig, int v) throws Exception, InvocationTargetException {
		Object[] argValues = {obj, clazz, fieldName, fieldSig, new Integer(v)};
		setIntField.invoke(null, argValues); //null because static
	}
*/


	/**------------------------------------------------------------------------
		* setIntFieldOpt (com.sun.corba.se.internal.io.IIOPInputStream
		* --------------
//private static native void setIntFieldOpt(Object o, long fieldID, int v);
		**/
/*
	private static final Method setIntFieldOpt;
	static {
		setIntFieldOpt = nativeCallInit(com.sun.corba.se.internal.io.IIOPInputStream.class,
			"setIntFieldOpt",
			new Class[] {Object.class, long.class, int.class});
	}
	public static void setIntFieldOpt(Object obj, ObjectStreamField f, int v)
throws Exception, InvocationTargetException {
		ObjectStreamField[] fs = {f};	//know element type because of field
		long[] primIDs = new long[1];	//we know only one field, and it's an Object
		long[] objIDs = new long[1];
		getFieldIDs(fs, primIDs, objIDs);
		Object[] argValues = {obj, new Long(primIDs[0]), new Integer(v)};	//note: prim (not obj)
		setIntFieldOpt.invoke(null, argValues); //null because static
	}
*/

/** --------------------
	* The first draft at initializing for allocateNewObject.
	**/
/*
	//Tuning: minimize object creation
		private final static Object[] argValues = new Object[2];
		private final static Class[] argTypes = {Class.class, Class.class};
//		private final static Method allocateNewObject;	//in static, so can catch
		static {
			Method m = null;	//in static, so can catch
			try {
				m = java.io.ObjectInputStream.class.getDeclaredMethod(
					"allocateNewObject", argTypes);
				m.setAccessible(true); //the magic paint
			} catch (Exception e) {
				System.err.println("'allocateNewObject' not found");
				e.printStackTrace();
				System.exit(0);
			}	
			allocateNewObject = m;	//ah!
//refactor control
		}
*/

}
