/** (c) Brendan Macmillan 2001, licensed under GNU's GPL 2 
    (Also available for non-GPL, commercial use, for fee) 
    Website: http://www.csse.monash.edu.au/~bren/JSX
    Website: http://freshmeat.net/projects/jsx
    List (can read without subscribing, need to join to post):
      http://groups.yahoo.com/group/JSX-ideas/messages
		Commercial licensing enquiries only:
			bren@mail.csse.monash.edu.au */

/** ===========================================================================
  * XMLDeserialize (JSX0.8.21).
	* ==============
	*
	* Creates the object described by an XML document.

	* Note that a basic limitation of Java is that an object cannot be created
	* unless the JVM has access to the class file defining it.  That is, Java
	* cannot *define* classes at runtime; it can only instantiate pre-defined
	* classes.
	* However, a class loader could be written to "read" in class files which
	* do not actually exist; thus giving Java the ability to define new
	* classes dynamically.  See also JavaClass (link on JSX homepage).

	* Please email me: for help, to suggest features, or to report bugs.
		bren@csse.monash.edu.au  <Brendan Macmillan>

	* ISSUES:
	*		What about: subclassing wrappers or containers?
	*		What if root object is a wrapper?  Or a field is a wrapper?
	
	**/


package JSX;
import java.lang.reflect.*;
import java.io.*;
import java.util.*;	//for Hashtable

class XMLDeserialize {

/** --------------------------------------------------------------------------
	* debug constants
	* ---------------
	**/
	private static final boolean VERSION_DEBUG = false;
	//private static final boolean VERSION_DEBUG = true;

	private static final boolean DEBUG_CUSTOM = false;
	//private static final boolean DEBUG_CUSTOM = true;

	private static final boolean DEBUG = false;
	//private static final boolean DEBUG = true;

	private static final boolean TESTCIRC = false;
	//private static final boolean TESTCIRC = true;

	private static final boolean ALIASDEBUG = false;
	//private static final boolean ALIASDEBUG = true;

	private static final boolean INTERNAL_DEBUG = false;
	//private static final boolean INTERNAL_DEBUG = true;

	private static final boolean DEBUG_CLASS = false;
	//private static final boolean DEBUG_CLASS = true;

	private static final boolean FINAL_DEBUG = false;
	//private static final boolean FINAL_DEBUG = true;

	static final boolean GETFIELD_DEBUG = false;
	//static final boolean GETFIELD_DEBUG = true;

	static final boolean READRESOLVE_DEBUG = false;
	//static final boolean READRESOLVE_DEBUG = true;

	static final boolean IBM_DEBUG = false;
	//static final boolean IBM_DEBUG = true;

	static final boolean KELLY_DEBUG = false;
	//static final boolean KELLY_DEBUG = true;

	private static final boolean SUPER_DEBUG = false;
	//private static final boolean SUPER_DEBUG = true;

	private static final boolean NESTED_DEBUG = false;	//part of super
	//private static final boolean NESTED_DEBUG = true;	//part of super

	private static final boolean NESTEDREADOBJECT_DEBUG = false;	//super
	//private static final boolean NESTEDREADOBJECT_DEBUG = true;	//super

	private static final boolean ADDATTR_NULL_DEBUG = false;
	//private static final boolean ADDATTR_NULL_DEBUG = true;

	private static final boolean KELLYSUPER_DEBUG = false;
	//private static final boolean KELLYSUPER_DEBUG = true;

	private static final boolean NOV_DEBUG = false;
	//private static final boolean NOV_DEBUG = true;

	//private static final boolean CFG_DEBUG = false;
	private static final boolean CFG_DEBUG = true;

/** --------------------------------------------------------------------------
	* Deserialization fields
	* ----------------------
	* NB: there are others scattered through the code, proximate to their use.
	**/
	private ParserXML p;

/** --------------------------------------------------------------------------
	* alias fields
	* ------------
	* Consider: extract as an object.  It has state, init, and API.
	* Consider: change name from 'aliasHash' to 'aliasTable'.
	**/
	private int aliasSerialNumber;	//NB: *not* static (cs serialization)
	private Hashtable aliasHash = new Hashtable();
		//private static final int ALIAS_INIT = -1;//so pre-increment in serializeObject
		//get INIT directly from XMLSerialize, so guaranteed to be synchronized.


	
	/**--------------------------------------------------------------------------
		* main
		* ----
		* Version information only; for test code, see JSXTest.java
		**/
	public static void main(String args[]) {
		System.err.println(XMLSerialize.VERSION);	
	}



/**==========================================================================
	* CONSTRUCTORS
	* ============
	**/
	/**--------------------------------------------------------------------------
		* XMLDeserialize(in) - core constructor
		* --------------
		* Core constructor (others call it)
		* init alias *only* here, or confuses re-enterant code (internal ser)
		* ALT: we could have an explicit null, instead of with/without CoNfig.
		**/
	XMLDeserialize(Reader in) {	
		reset();	//whenever create, always reset alias table.
		p = new ParserXML(in);
	}

	private Config cfg;
	XMLDeserialize(Reader in, Config cfg) {	
		this.cfg = cfg;
		reset();	//whenever create, always reset alias table.
		p = new ParserXML(in);
	}

	//--------------------------------------------------------------------------
	/** XMLDeserialize()
		* --------------
		* Default to System.in
		**/
	XMLDeserialize() {			//default to System.in
		this(new InputStreamReader(System.in));
	}



/**==========================================================================
	* METHODS
	* =======
	**/

		//use explicit subclass, so we can access our "putObj()" method
	GetFieldImpl currentGetField = new GetFieldImpl();	//NESTED FIELDS - needed?
	
		//this is set, ready for ObjIn to grab it
	static final class GetFieldImpl extends java.io.ObjectInputStream.GetField {
		//interal - *not* part of public API
		HashMap objMap = new HashMap();	//need to distinguish obj and prim?
		public String toString() {
			return "GetFieldImpl state: "+objMap;
		}
		GetFieldImpl() {
		}
		void putObj(String name, Object value) {
			if (GETFIELD_DEBUG) System.err.println("putObj("+name+", "+value+");");
			objMap.put(name, value);
		}

		public ObjectStreamClass getObjectStreamClass() {
			return null;		//DEFER: this will be tricky to implement... or a
			//convenience, to encourage me to implement it closer to their way?
		}
		public boolean defaulted(String name)
			throws IOException, IllegalArgumentException {
			return !objMap.containsKey(name);	//if containsKey, then not default...
		}
		public boolean get(String name, boolean defvalue) 
				throws IOException, IllegalArgumentException {
			if (objMap.containsKey(name)) 
				return ParseUtilities.parseBoolean((String)objMap.get(name));	//how to tell if default?
			else
				return defvalue;
		}
		public char get(String name, char defvalue) 
				throws IOException, IllegalArgumentException {
			if (objMap.containsKey(name)) 
				return ((String)objMap.get(name)).charAt(0);
			else
				return defvalue;
		}
		public byte get(String name, byte defvalue) 
				throws IOException, IllegalArgumentException {
			if (objMap.containsKey(name)) 
				return Byte.parseByte((String)objMap.get(name));
			else
				return defvalue;
		}
		public short get(String name, short defvalue) 
				throws IOException, IllegalArgumentException {
			if (objMap.containsKey(name)) 
				return Short.parseShort((String)objMap.get(name));
			else
				return defvalue;
		}
		public int get(String name, int defvalue) 
				throws IOException, IllegalArgumentException {
			if (objMap.containsKey(name)) 
				return Integer.parseInt((String)objMap.get(name));
			else
				return defvalue;
		}
		public long get(String name, long defvalue) 
				throws IOException, IllegalArgumentException {
			if (objMap.containsKey(name)) 
				return Long.parseLong((String)objMap.get(name));
			else
				return defvalue;
		}
		public float get(String name, float defvalue) 
				throws IOException, IllegalArgumentException {
			if (objMap.containsKey(name)) 
				return ParseUtilities.parseFloat((String)objMap.get(name));
			else
				return defvalue;
		}
		public double get(String name, double defvalue) 
				throws IOException, IllegalArgumentException {
			if (objMap.containsKey(name)) 
				return ParseUtilities.parseDouble((String)objMap.get(name));
			else
				return defvalue;
		}
		public Object get(String name, Object defvalue) 
				throws IOException, IllegalArgumentException {
			if (objMap.containsKey(name)) 
				return objMap.get(name);
			else
				return defvalue;
		}
	}



	int invocationCount = 0;
	//--------------------------------------------------------------------------
	/** deserialize()
		* -----------
		* Entry point.
		* NB: invocationCount distinguisges between first time callers, and
		* calls from internal serialization.  This allows both old and new
		* API's to work.
		**/
	String jsxVersion;		//what will happen to this information?
			//After decoding information, need to interpret it somewhere.  Where?
	boolean superVersion = false;	//default (for now)

/*
	class Header {
		public Header(String target) { this.target = target; }
		final String target;
		final String encoding;
		final String version;	//dangerous? if move to 1.0.1? Eg for XML, it is
			//"1.0"...	tests for exact equivalence?  Best to use Strings..
	}
	//Header xmlHeader;
	//Header jsxHeader;
*/
	Object deserialize() throws ParserXML.ExceptionXML,
			                               ClassNotFoundException,
			                               //NoSuchFieldException,
                                     IOException {
                                     //Exception {
		if (INTERNAL_DEBUG) System.err.println("RETRIEVE optional data object: "+objStoreIndex);
		if (invocationCount==0) {
			reset();
			invocationCount++;
			ParserXML.Tag tag = p.readTag();
			if (tag==null) throw new EOFException("No XML found in input");		//no more objects left
			while (tag.isPI) { //many PI's to process (eg ?xml, ?jsx)
				//Polymorphic information storage:
				//	Header currentHeader = new Header(tag.name);
				// if (tag.name.equals(XMLSerialize.JSX_HEADER_TARGET)) {//"factory if"
				//		jsxHeader = currentHeader;	//set this field to it.
				// }
				//	else if (tag.name.equals(XMLSerialize.XML_HEADER_TARGET)) {
				//		xmlHeader = currentHeader;	//set this field to it.
				//	}
				//	else {	//unrecognized target: info is lost.
				//	}
				//		
				if (VERSION_DEBUG) System.err.println("start PI: "+ tag);
				ParserXML.Attr attr;
				while ( !(attr=p.readAttr()).isEndPI) {	//attrs (eg version, encoding)
					if (attr.isEnd)
						throw new IOException("PI should end with '?>', not '/>'");
					if (attr.name.equals("version")) {	//this is kinda cool ;-0 oh yeah
						if (VERSION_DEBUG) System.err.println("\tversion: "+attr.value);
						//(1). set version
								//	currentHeader.version = attr.value;
						if (tag.name.equals(XMLSerialize.JSX_HEADER_TARGET)) {
							jsxVersion = attr.value;
							if (jsxVersion.equals("1")) {
								superVersion = false;
							} else if (jsxVersion.equals("2")) {
								superVersion = true;
							} else {
								throw new IOException("Unknown jsx version=\""+jsxVersion+"\".  You need to upgrade to read this later version");
							}
						}
					} else if (attr.name.equals("encoding")) {
						if (VERSION_DEBUG) System.err.println("\tencoding: "+attr.value);
						//(2). set encoding
						// currentHeader.encoding = attr.value;
					} else {
						if (VERSION_DEBUG) System.err.println("\tunknown attr: "+attr+" in PI: "+tag);
						//ignore unknown options.
					}
				}
				if (VERSION_DEBUG) System.err.println("end PI: ?>");
				tag = p.readTag();	//set up for root (or another PI)
					//NB: error if null (root *must* follow PI)
			}
		//Version info is needed where?
			//(1). Pass it back to the caller? (ie ObjIn)
			// - perhaps easiest to register a call-back, listener: std architecture
			//(2). Set it directly with the registered back-reference?
			//(3). Set it directly with a setVersion(version) on the back-reference?
			//NOTE: No point planning too much.  Wait for JAXP - just do it.
			//(4). We could send it back to ObjIn, along with ParserXML p, which
			//could then easily be plugged into another XMLDeserialize.  Note that
			//this won't help for version changes to the *parser* - but that should
			//be OK.
			//(5). Perhaps the key thing is to swap streams with JAXP... actually,
			//that will be the one reading the stream, when it finds it needs to
			//swap.  At that point, easiest might be to reset the stream to the
			//start, and start all over again.

			if (!tag.start)
				throw new ParserXML.ExceptionXML("Expected start tag, got \"</" +tag.name+ ">\"");
			//Object o = createObject(tag.name, null);	//no parent
			Object o = createObject(tag, null);	//no parent //10 Apr 2001
			if (TESTCIRC)
				System.err.println("Count was: " + (aliasSerialNumber-XMLSerialize.ALIAS_INIT));
				if (ALIASDEBUG) System.err.println(aliasHash);
			invocationCount--;
			return o;	//separate statement into 2 lines so we can print the count
		}
		else {//ie. if from a customization
			if (superVersion) {
				defROSubstitute();	//eat any header stuff, if defRO not called
				ParserXML.Tag tag = p.readTag();
				if (NOV_DEBUG)
					System.err.println("readObject: just read a "+ tag);
				return createObject(tag, null);
			}
			else
				return getObj();
		}
	}

	void defROSubstitute() throws IOException {
		try {
			if (NOV_DEBUG) System.err.println("entering defROSubstitute:");
			if (defROCalledStack.get(defROCalledStack.size()-1)==null) {
				if (NOV_DEBUG) System.err.println("\tfor first time");
				defROCalledStack.set(defROCalledStack.size()-1, Boolean.TRUE);
				defaultReadObject_ver2();	//to just eat!
	/*
					//we don't actually use "Boolean true" - but nice description
			//need to read the header stuff that defRO would normally read:
				ParserXML.Attr[] addAttrTmp = defaultReadObject_attrs_in;
						//stack unnec, as use immediately
				addPrim(addAttrTmp, null, false);	//should not return a name
				ParserXML.Tag tag = p.readTag();	//should be <opt-data> section
  			checkOpenTag(XMLSerialize.OPT_DATA_TOKEN, tag);
				System.err.println("reading <opt-data> from within readObject()");
				addPrim(addName(), null, false);	//just clear end of opt tag
				defROStack.set(defROStack.size()-1, tag);	//mimic defRO
	*/
			}
		} catch (ClassNotFoundException e) {
			System.err.println("unexpected default fields - that's OK in itself,"
				+ "the problem is: "+e);
			e.printStackTrace();
		}
	}

	//NB: same as XMLSerialize
	void reset() {
		if (invocationCount!=0)
			throw new Error("Deserialiize: Attempt to reset alias table mid-way.");
		initAlias();
	}
	void close() {	p.close(); }



	/** =========================================================================
		* alias nascent-object
		* --------------------
		* NB: init, fields and API: extract into a class.
		* Should move this into a separate, pluggable class, so that
		* (1). the implementation can be changed easily
		* (2). a co-developer can work on it in isolation
		* (3). it can be recompiled separately - faster recompilation
		* (4). it can be reused elsewhere
		* (5). a different implementation can be plugged in (including a subclass)
		*
		* All are aspects of "divide and conquer" - the danger is to divide 
		* overzealously, in an inappropriate way, which counterproductively
		* increases complexity.  However, when a division is clear, it really
		* should be done.
		* The reluctance to do it is that the change may introduce bugs.  It is
		* work, and uncertainty, and fear of opening a can of worms - better to
		* let sleeping dogs lie.
		* And yet, because it has been used as it were a separate object, it
		* should not be too hard.  And if there are ripples in the force, it is
		* good to find them and tame them, rather than fear them.
		*
		* It would also be good to have an "alias" API - javadocs could produce
		* this automatically.  Makes it easier to work with.  Looks more
		* professional (to employers, licensees), and more welcoming to 
		* co-developers.   Reduces barriers to contribution.  Increases chances
		* of patches being offered.
		**/ 
	private void initAlias() {
		aliasSerialNumber = XMLSerialize.ALIAS_INIT;
		aliasHash.clear();
	}
	/**--------------------------------------------------------------------------
		* put Alias
		* --------
		*	inline = !serialized;
		* if (inline)
		*		//add to hash
		*	else
		*		//write as alias
		*	return (inline);		//to tell caller to serialize it inline!
		*		//else, caller leaves the ref writing up to this routine
		**/
	//private static boolean putAlias(Object o, String fieldName, boolean isStringAttr, String indent) {

	/**--------------------------------------------------------------------------
		* getAlias
		* --------
		* Much of this is duplicated code from addAttr().  This should be
		*	rationalized and refactored at some point.
		**/
	boolean aliasNameMissing;	//clumsy message passing
	private NamedObject getAlias(ParserXML.Attr[] addAttrTmp, String flag, Object parent) throws
			//NoSuchFieldException,
			IOException {
		aliasNameMissing = false;
		ParserXML.Attr attr = addAttrTmp[0];	//now passed in above
		String fieldName = null;
		String aliasName = null;
		if (flag.equals(XMLSerialize.ALIAS_TAG_TOKEN)) {	//parse obj-name and name
			//while ( !(attr=p.readAttr()).isEnd) 
			if (!attr.isEnd) {
				do {
					if (attr.name.equals(XMLSerialize.NAME_TOKEN)) {
						if (fieldName!=null) 
							throw new ParserXML.ExceptionXML("just one name field");
						fieldName = attr.value;
					}
					else if (attr.name.equals(XMLSerialize.ALIAS_ATTR_TOKEN)) {
						if (aliasName!=null)
							throw new ParserXML.ExceptionXML("just one alias name");
						aliasName = attr.value;		//NB: won't flag duplicates
					} else
							throw new ParserXML.ExceptionXML("unknown attribute: "+attr.name);
				} while (!(attr=p.readAttr()).isEnd);
			}//if
 			Object me = aliasHash.get(aliasName);
			if (parent!=null) {	//ie, if not an orphan object (as in Vector, Array)
				if (fieldName==null)
					aliasNameMissing = true;	//need to defer it, so context is correct
				else {
					//try {
						currentGetField.putObj(fieldName, me);
						Field f = getAllField(parent.getClass(), fieldName);
						setFinal(f, parent, me);
						//f.set(parent, me);
					//} catch (NoSuchFieldException e) {
					//}
				}
			}
				//System.err.println("alias " +aliasName+ ":");
				//System.err.println("me: " +me);
 			return new NamedObject(fieldName, me, !aliasNameMissing);
		}
		else
			return null;	//indicates that the tag was not an alias
	}

	private static void setFinal(Field f, Object parent, Object value) {
	 if (f==null) return;
	 try {
		if (FINAL_DEBUG) {
			System.err.println("setFinal");
			System.err.println(f);
			System.err.println(parent);
			System.err.println(value);
		}
		if (Modifier.isFinal(f.getModifiers())) {	//if final...
			//object and primitive need osf.  We simulate a non-private constructor:
					/*		ObjectStreamField(Field field) {
					*			this(field.getName(), field.getType());
					*			this.field = field;
					*		} */
			//How can this be achieved more simply?
			ObjectStreamField osf = new ObjectStreamField(f.getName(), f.getType());
			Field osfField = ObjectStreamField.class.getDeclaredField("field");
			osfField.setAccessible(true);
			osfField.set(osf, f);
			//The previous approach was:
				//ObjectStreamClass osc = ObjectStreamClass.lookup(parent.getClass());
				//System.err.println(parent.getClass());
				//System.err.println(osc);
				//System.err.println(f.getName());
				//ObjectStreamField osf = osc.getField(f.getName());
			if (FINAL_DEBUG) {
				System.err.println(osf);
				System.err.println("about to set...");
			}
			if (f.getType().isPrimitive()) {	//if prim...
				MagicClass.setPrimitiveFieldValues(parent, osf, value);
				if (FINAL_DEBUG) System.err.println("...set primitive final field!");
			} else { //if obj...
				MagicClass.setObjectFieldValue(parent, osf, f.getType(), value);
				if (FINAL_DEBUG) System.err.println("...set object!");
			}
		} else {	//if non-final:
			//do with reflection
			//Future: tigher integration with above (array based) is faster.
			f.set(parent, value);	//<--- this doesn't work
		}
		if (FINAL_DEBUG) System.err.println("finished setFinal");
	 } catch (IllegalAccessException e) {
			//this may indicate permissions need to be changed: should propagate?
	 } catch (NoSuchFieldException e) {
			//Special case: ObjectStreamField should have a field named "field".
	 }
	}

	/** ------------------------------------------------------------------------
		* putAlias
		* --------
		* Change to return the alias - this should be a "back-compatible
		* refactoring" for callers (tho all return statements need to be changed),
		* and it allows it to be stored, and used to later update it (for
		* readResolve()).
		**/
	private String putAlias(Object o) {
		if (o==null) return null;	//a flag: null not a valid alias - but...
			//better to throw an exception?
		++aliasSerialNumber;	//pre-inc (use this syntax to highlight it)
		String alias = aliasSerialNumber+"";
		aliasHash.put(alias, o);	//add (NB: inverse of serialize)
			//put() returns the previous key - it *shouldn't* have had one (ie null)
		if (TESTCIRC) System.err.println("inline " +aliasSerialNumber+ ":");
		return alias;
/*
		Object alias;	//will be String, but save the conversion
		if ( (alias = aliasHash.get(o))==null) {
			++aliasSerialNumber;	//pre-inc (use this syntax to highlight it)
			aliasHash.put(o, aliasSerialNumber+"");	//add (convert to String)
				//System.err.println(indent +"inline " +aliasSerialNumber+ ":");
			//Need to:
			//	indicate "inline" somehow.
			//	write as inline (same as now).
			return true;	//yes, should serialize the object inline
		}
		else
		{
			if (isStringAttr) {
				printAttr(fieldName + STRING_ALIAS_TOKEN, alias+"");
			}
			else {
					//System.err.println(indent+ "alias " +alias+ ":");
				printOpenTag(ALIAS_TAG_TOKEN, indent);
				if (fieldName!=null) printAttr(NAME_TOKEN, fieldName);	//nice to reuse
				printAttr(ALIAS_ATTR_TOKEN, alias+"");
				out.println("/>");		//empty tag - under control of caller
			}
			return false;	//no, don't serialize inline; we (or you?) write the alias
		}
*/
	}

	/** ------------------------------------------------------------------------
		* updateAlias
		* -----------
		* for readResolve() - easy to implement, because Hashtable already
		* allows for updates
		**/
	private String updateAlias(Object o, String alias) throws IOException {
		if (o==null) return null;	//a flag: null not a valid alias - but...
			//better to throw an exception?
		if (aliasHash.put(alias, o)==null)	//add (NB: inverse of serialize)
			throw new IOException("update called, but no existing alias");
			//this returns the previous key - it *should* have had one.
		return alias;
	}


	private ObjIn oisSubclass;	//we can call protected, because same package
		//the problem before was with ObjectInputStream - not same (JSX) package!
	static final Class[] OIS_ARGS = {ObjectInputStream.class};
	Object[] readObjectArglist;
	/**--------------------------------------------------------------------------
		* setArg()
		* -------
		* Consider: putting this in constructor; or caller in same package or same
		* class.
		**/
	void setArg(Object[] argList) {
		readObjectArglist = argList;
		oisSubclass = (ObjIn) argList[0];	//kill two birds with one stone
	}
	Vector primStore = null; //Note: not yet set!  Consider: best place to define?
	int primStoreIndex = -1;	//fail fast
	/**--------------------------------------------------------------------------
		* getPrimString()
		* ---------------
		* Called by readInt(), readFloat(), etc to get the String that was
		* stored by addAttr
		**/
	String getPrimString() throws IOException {	//note: we *know* what it is meant to be...
		if (superVersion) {		//what to do here?
			defROSubstitute();	//eat any header stuff, if defRO not called
			return deserializePrimOptData();
		} else {
			return (String) primStore.get(primStoreIndex++);
		}
	}

	//			<prim-opt-data valueOf="___"/>  <---- only this version is expected
	//			<prim-opt-data valueOf="___">   <---- but we can handle this one too
	//			</prim-opt-data>
	private String deserializePrimOptData() throws IOException {
			//	<prim-opt-data..
		ParserXML.Tag tag = p.readTag();
		checkOpenTag(XMLSerialize.OPT_PRIM_DATA_TOKEN, tag);
			// 		            ..valueOf="___"..
			//															 ../>
		ParserXML.Attr attr;
		String value = null;
		while (!(attr=p.readAttr()).isEnd) { //slurp the lot... so can ignore some
			if (attr.name.equals(XMLSerialize.VALUE_TOKEN))
				value = attr.value;
		}
		if (value==null)
			throw new ParserXML.ExceptionXML("Expected: "+XMLSerialize.VALUE_TOKEN);

		if (!attr.emptyTag) 	//the final one
			//	</prim-opt-data>
			checkCloseTag(XMLSerialize.OPT_PRIM_DATA_TOKEN, p.readTag());	//cool!
		return value;	//my work here is done...
	}


	Vector objStore = new Vector(); //Just to debug superVersion
			//Consider: best place to define?
	int objStoreIndex;
	/**--------------------------------------------------------------------------
		* getObj()
		* --------
		* Called by readObject(), when it is a from a customization (not the
		* top-level call).  The Vector is populated by createObject.
		**/
	Object getObj() {
		return objStore.get(objStoreIndex++);
	}

	/**--------------------------------------------------------------------------
		* NamedObject - message object (to return more than one thing)
		* -----------
		* Defn: Wraps an "object" and its (field)"name".
		* Used by: createObject. null, alias, array etc routines return it.
		*
		* Future: We return a message object.  Object creation is inefficient.
		* Reuse a static object - a global variable with cleaner semantics (usage).
		* That is, set global/static object fields, and return it.  Problem is that
		* not restricted to this use - others could still access it as global -
		* unless the orginator of it was in a separate class.  An advantage of
		* small modules.
		**/
	class NamedObject {
		String name;
		Object object;	//better to not use abbreviations
		boolean named;
		NamedObject(String name, Object object, boolean named) {
			this.name = name;
			this.object = object;
			this.named = named;
		}
	}


	/**--------------------------------------------------------------------------
		* createObject()
		* ------------
		* Returns the deserialized object (caller needs to cast at runtime)
		* The object created is "me".
		* USAGE: the open tag is assumed to have been just read
		* LOGIC: A new object is instantiated: "me".  Attr and elements are
		* added to me.  This new object is then added to its parent.  NB: that
		* the name for me (given by parents) is only discoverable by reading the
		* attr; which is done in "addAttr" (second arg is boolean for isArray)
		* NB: a null object may still have a name.
		* NOTE: tag problem for lengths - not actually a field.
		*
		* Issues are:
		*	 - custom data? Unnamed fields.
		*  - Strings - attr or tags? strings in Object fields.
		*
		* parentClassFrame
		* ----------------
		* Parent class frame is needed when we want to assign this tag to its
		* parent - it's a namespacing thing.  We can rely on the shadow namespacer
		* as scaffolding, without actually using this one for now.
		**/
	//private Object createObject(ParserXML.Tag tag, Object parent, Class parentClassFrame)


	private Object createObject(ParserXML.Tag tag, Object parent)
			throws IOException,
						 //NoSuchFieldException,
			       ClassNotFoundException {
			       //Exception 
		ParserXML.Attr[] addAttrTmp = addName();  //Too early for special cases?
		return createObject(tag, addAttrTmp, parent);
		//return createObject(tag, addName(), parent);	//consider this...
	}

	private Object createObject(ParserXML.Tag tag, ParserXML.Attr[] addAttrTmp, Object parent)
			throws IOException, ClassNotFoundException {
		String elementName = tag.name;	//loosely coupled argument passing
		if (KELLY_DEBUG)
			System.err.println("adding "+elementName);

		//Notes on "sub-class" field:
		//---------------------------
		//Theory: For subclasses, the "sub-class" field is actually the same
		//object, just with a different classFrame through which it is viewed
		//(revealing different fields, and also a different readObject() to call).
		//Thus, for a subclass tag, its "parent" is actually the same object.
		//Implementation:  As a hack, instead of explicitly checking if the present
		//tag is a subclass field, we set it as a default - if the tag we are in
		//turns out to *not* be a subclass, these default values will be written
		//over.
		//For special-case tags: A subclass can never be null, or an alias (each
		//instance occurs exactly once.  It cannot be String or Class, because
		//these are final.  However, it could be a Vector or Hashtable - not
		//sure how these will be coped with.
		Object me = parent;	//default; me is overwritten if not a subclass.
			//Needed? No:
			// - if subclass, this tag sets a new classFrame (the "next" one);
			// - if not subclass (ie new header), will set an entirely new frame.
		//Class meClassFrame = parentClassFrame;	//again, default/overwritten

		boolean nameMissing = false;	//named assumed to be present
		NamedObject namedObject;	//used in many sections below:

//NULL
		if (elementName.equals(XMLSerialize.NULL_TOKEN)) {
			//namedObject = deserializeNull(parent); //code dup for array and alias
			namedObject = deserializeNull(addAttrTmp, parent); //code dup for array and alias
			me = namedObject.object;	//it will be null
			nameMissing = !namedObject.named;	//this is the info we need
			if (nameMissing) {	//only consider objStore, if name is missing...
				if (parent!=null) {	//if not an orphan, for all these...
					if (parent.getClass()!=Vector.class &&			//not vector
							parent.getClass()!=Hashtable.class &&		//not hashtable
				 			!parent.getClass().isArray()) {					//not array
						objStore.add(me);
					}
				}
			}
			//NOTE: a bug was fixed in this refactoring: before, such a string was
			//added to objStore, regardless of whether named or not.  However, this
			//wouldn't confuse any custom objects, because this extra one would just
			//be tacked on the end.
			return me;
		}

//ARRAY
		else if (elementName.startsWith(XMLSerialize.ARRAY_TOKEN)) {
			//namedObject = createArray(elementName, parent);
			namedObject = createArray(addAttrTmp, elementName, parent);
			me = namedObject.object;
			nameMissing = !namedObject.named;	//this is the info we need
			if (nameMissing) {	//only consider objStore, if name is missing...
				if (parent!=null) {	//if not an orphan, for all these...
					if (parent!=null && //not root (MEANT to be "not orphan"?)
							parent.getClass()!=Hashtable.class &&
							parent.getClass()!=Vector.class &&
							!parent.getClass().isArray()) {
						objStore.add(me);	//old bug: no objStore exists.
					}
				}
			}
			//Only put in objStore if custom data. That is, if an "unnamed field":
			//parent is a classInstance and object has no name.
			return me;		//Limit interaction with rest of routine.
			//NOTE createArray() reads close tag, to completely parses the element.
			//validation of nested of array types is left to Java.
		}

//ALIAS
		else if ( (namedObject=getAlias(addAttrTmp, elementName, parent))!=null ) {
			me = namedObject.object;
			nameMissing = !namedObject.named; //if named, plugged inside already
			if (nameMissing) {	//only consider objStore, if name is missing...
				if (parent!=null) {	//if not an orphan, for all these...
					if (parent.getClass()!=Vector.class &&			//not vector
							parent.getClass()!=Hashtable.class &&		//not hashtable
			 				!parent.getClass().isArray()) {					//not array
							objStore.add(me);
					}
				}
			}
			return me;
		}

//BINARY DATA (similar design to Class and String)
		else if (elementName.equals(XMLSerialize.BINARY_DATA_TOKEN)) {
			namedObject = deserializeBinary(addAttrTmp, tag, parent);
				//(1). sets the field, if name is null.
				//(2). putAlias()
			nameMissing = !namedObject.named; //if named, already plugged
			me = namedObject.object;
			if (parent!=null) {	//if not an orphan, for all these...
				if (nameMissing) {	//only consider objStore, if name is missing...
					if (parent.getClass()!=Vector.class &&			//not vector
						parent.getClass()!=Hashtable.class &&		//not hashtable
						!parent.getClass().isArray()) {						//not array
						objStore.add(me);	//NOTE: we store it as an Object, not prim!
					}
				}
			}
			return me;
		}

//STRING
		else if (elementName.equals("java.lang.String")) {
				//EXAMPLE:  Object eg = "a string";  //<- String as named object
				//un-named: element of Hash or Vect, array of Object(?), root
			namedObject = deserializeWrapper(addAttrTmp, tag, parent);
				//(1). sets the field, if name is null.
				//(2). putAlias()
			nameMissing = !namedObject.named; //if named, already plugged
			me = namedObject.object;
			if (parent!=null) {	//if not an orphan, for all these...
				if (nameMissing) {	//only consider objStore, if name is missing...
					if (parent.getClass()!=Vector.class &&			//not vector
						parent.getClass()!=Hashtable.class &&		//not hashtable
						!parent.getClass().isArray()) {						//not array
						objStore.add(me);	//NOTE: we store it as an Object, not prim!
					}
				}
			}
			return me;
		}

//CLASS
		else if (elementName.equals("java.lang.Class")) {
			if (DEBUG_CLASS)
				System.err.println("Got a Class tag");
			namedObject = deserializeClass(addAttrTmp, tag, parent);
				//(1). sets the field, if name is null.
				//(2). putAlias()
			nameMissing = !namedObject.named; //if named, already plugged
			me = namedObject.object;
			if (parent!=null) {	//if not an orphan, for all these...
				if (nameMissing) {	//only consider objStore, if name is missing...
					if (parent.getClass()!=Vector.class &&			//not vector
						parent.getClass()!=Hashtable.class &&		//not hashtable
						!parent.getClass().isArray()) {						//not array
						objStore.add(me);	//NOTE: we store it as an Object, not prim!
					}
				}
			}
			if (DEBUG_CLASS) {
				System.err.println(">>>>start>>>>>");
				System.err.println(me);
				try {
					System.err.println( ((Class)me).newInstance());
					new JSX.ObjOut(System.err).writeObject( ((Class)me).newInstance());
				} catch (InstantiationException e) {
					System.err.println(e);
					e.printStackTrace();	//OK, because within DEBUG block
				} catch (IllegalAccessException e) {
					System.err.println("Should never get here - it means that we couldn't create a Hashtable with the no-arg constructor");
					e.printStackTrace();
				}
				System.err.println("<<<<end<<<<");
			}
			return me;
		}

//VECTOR
		else if (elementName.equals("java.util.Vector")) {
			Class clazz = Class.forName(ParseUtilities.descapeDollar(elementName)); //2 lines, for debugging
			me = null;	//why not just make one, if we are so sure it is OK?
			try {
				me = clazz.newInstance();
			} catch (InstantiationException e) {
				if (DEBUG) {
					System.err.println("Should never get here - it means that we couldn't create a Hashtable with the no-arg constructor");
					e.printStackTrace();
				}
			} catch (IllegalAccessException e) {
				if (DEBUG) {
					System.err.println("Should never get here - it means that we couldn't create a Hashtable with the no-arg constructor");
					e.printStackTrace();
				}
			}
			ParserXML.Attr name = null;
			name = addPrim(addAttrTmp, me, false);//gets NAME_TOKEN - should be no primitive data
	 		nameMissing = name.nameMissing;
			 if (parent!=null) { //can't plug if no parent (vect, hash, array do sep)
				//(1). BUFFER (CUSTOM DATA)
	 			if (nameMissing) {//no name
					if (parent.getClass()!=Vector.class &&			//not vector
							parent.getClass()!=Hashtable.class &&		//not hashtable
							!parent.getClass().isArray()) {						//not array
						objStore.add(me);	//store custom *Object* in buffer
					}
				}
				else {	//if named...
				//(2). INTO CLASS AS FIELD
					//try {
						currentGetField.putObj(name.value, me);	//prim/obj?
						Field f = getAllField(parent.getClass(), name.value);
						setFinal(f, parent, me);
						//f.set(parent, me);
					//} catch (NoSuchFieldException e) {
					//}
				}
			}
			putAlias(me);	//to sync with ser: do before putAlias for elements
	 		if (!name.emptyTag) {	//<java.util.Vector/> means no elements to add
				addVectorElements((Vector) me);
			}
			return me;	//note: could return null, if already plugged with name?
		}

//HASHTABLE
		else if (elementName.equals("java.util.Hashtable")) {
			Class clazz = Class.forName(ParseUtilities.descapeDollar(elementName)); //2 lines, for debugging
			me = null;	//why not just make one, if we are so sure it is OK?
			try {
			me = clazz.newInstance();
			} catch (InstantiationException e) {
				if (DEBUG) {
					System.err.println("Should never get here - it means that we couldn't create a Hashtable with the no-arg constructor");
					e.printStackTrace();	
				}
			} catch (IllegalAccessException e) {
				if (DEBUG) {
					System.err.println("Should never get here - it means that we couldn't create a Hashtable with the no-arg constructor");
					e.printStackTrace();
				}
			}
			ParserXML.Attr name = null;
			name = addPrim(addAttrTmp, me, false);//gets NAME_TOKEN - should be no primitive data
	 		nameMissing = name.nameMissing;
			if (parent!=null) { //can't plug if no parent (vect, hash, array do sep)
				//(1). BUFFER (CUSTOM DATA)
	 			if (nameMissing) {//no name
					if (parent.getClass()!=Vector.class &&			//not vector
							parent.getClass()!=Hashtable.class &&		//not hashtable
							!parent.getClass().isArray()) {						//not array
						objStore.add(me);	//store custom *Object* in buffer
					}
				}
				else {	//if named...
				//(2). INTO CLASS AS FIELD
					//try {
						currentGetField.putObj(name.value, me);	//prim/obj?
						Field f = getAllField(parent.getClass(), name.value);
						setFinal(f, parent, me);
						//f.set(parent, me);
					//} catch (NoSuchFieldException e) {
					//}
				}
			}
			putAlias(me);	//to sync with ser: do before putAlias for elements
	 		if (!name.emptyTag) {	//<java.util.Hashtable/> means no elements to add
				addHashtableElements((Hashtable) me);
			}
			return me;	//note: could return null, if already plugged with name?
		}


//CLASSINSTANCE (the entire remainder of routine)
//-------------
// ---just written: 2 Aug 2001---
//proposed change - just a switch in order, and move some to default() method.
//	- create
//	- foreach class, call readObject
//	- else default
//		- default is addAttr, plug, addTag
//	NB: need addAttr for name, to decide how to plug.
//	NB: always call default *exactly* once
//	NB: maybe problems with stores (for custom data)
//	NB: this fix only works if only one class in the inheritance path
//		has a readObject() that relies on the defaultReadObject() order.
//		- and fortunately, this does seem to be the case.
		else {
			//ParserXML.Attr[] addAttrTmp = addName();//head of old "addAttr()"
			ParserXML.Attr name = addAttrTmp[1];
			boolean subclass = false;	//ie "not header" the default
			if ( name!=null && name.value.equals(XMLSerialize.SUPER_TOKEN) ) {
				subclass = true;
			}
			String alias = null; //store for (possible) readResolve
			Class meClassFrame;  //we may need to set this for first call?

		 if (!subclass) {
			Class clazz = null;
			//clazz = Class.forName(ParseUtilities.descapeDollar(elementName));	//note: code dup with above
			//System.err.println("forName: "+clazz+", "+clazz.getClassLoader());

			//System.err.println("XMLDeserialize: About to load class");

			//front-end to for java.io.ObjectInputStream.resolveClass()
				//construct an osc object with "name" of desired class to load.
			String escapedName = ParseUtilities.descapeDollar(elementName);
			ObjectStreamClass osc = null;
			try {
				osc = (ObjectStreamClass) MagicClass.newInstance(ObjectStreamClass.class);
				Field oscNameField = ObjectStreamClass.class.getDeclaredField("name");
				oscNameField.setAccessible(true);	//"name" is private
				//System.err.println("osc: "+osc);
				setFinal(oscNameField, osc, escapedName);
				//System.err.println("osc: "+osc);
			} catch (InvocationTargetException e) {
					//a crucial exception
				Throwable thrown = e.getTargetException();
				if (thrown instanceof ClassNotFoundException)
					throw (ClassNotFoundException) thrown;	//rethrow it
				else if (thrown instanceof IOException)
					throw (IOException) thrown;	//rethrow it
				else {
					System.err.println("\nWrapper Exception:");
					e.printStackTrace();
					System.err.println("\nException wrapped up:");
					thrown.printStackTrace();
				}
			} catch (NoSuchFieldException e) {
					e.printStackTrace();
					//Special case: ObjectStreamField should have a field named "name".
					//...but this does rely on the private implementation...
			} catch (IllegalAccessException e) {
					e.printStackTrace();
			}
			clazz = oisSubclass.resolveClass(osc);	//done!
			if (clazz==null)
				throw new ClassNotFoundException("Could not load "+escapedName);
			//System.err.println("resolveClass: "+clazz+", "+clazz.getClassLoader());

/*
		try {
			Method latest = ObjectInputStream.class.getDeclaredMethod("latestUserDefinedLoader", new Class[0]);
			latest.setAccessible(true);	//private
			ClassLoader loader = (ClassLoader)latest.invoke(null, null);
			System.err.println(" --- latest loader --->  "+loader);
			clazz = Class.forName(elementName, false, loader);
			System.err.println("XMLDeserialize latest: "+clazz+", "+clazz.getClassLoader());
		} catch (Exception e) {e.printStackTrace();}
			System.err.println("XMLDeserialize call to Latest.get");
		classLoading.Latest.get();

			System.err.println("XMLDeserialize IN LINE CODE in line code");
	//does it make a difference being right here?  YES!! crazy...
    try {
      java.lang.reflect.Method latest = java.io.ObjectInputStream.class.getDeclaredMethod("latestUserDefinedLoader", null);
  //d("latestUserDefinedLoader", new Class[0]);
      latest.setAccessible(true); //private
      ClassLoader loader = (ClassLoader)latest.invoke(null, null);
      System.err.println("latest loader: "+loader);
    } catch (Exception e) {
      e.printStackTrace();
    }
*/
/*
			clazz = Class.forName(elementName, false, ObjIn.cl);	//cheat
*/


			me = null;	//is there a better way to do this?
			try {
				if (Externalizable.class.isAssignableFrom(clazz)) {
					try {
						Constructor cons = clazz.getConstructor(new Class[0]);
						//cons.setAccessible(true);	//that's why we can't clazz.newInstance()
						me = cons.newInstance(new Object[0]);	//Externalizable requires no-arg constructor
					} catch (NoSuchMethodException e) {
						throw new InvalidClassException("Missing public no-arg constructor for class "+ clazz.getName());
					} catch (InstantiationException e) {
						throw new InvalidClassException("An interface or abstract class "+ clazz.getName());
					}
				} else {
					me = MagicClass.newInstance(clazz);	//use anon
					if (me==null)
						throw new ClassNotFoundException("MagicClass.newInstance() failed to create a "+clazz+", from the name "+escapedName);
				}
			} catch (InvocationTargetException e) {
					//a crucial exception
				Throwable thrown = e.getTargetException();
				if (thrown instanceof ClassNotFoundException)
					throw (ClassNotFoundException) thrown;	//rethrow it
				else if (thrown instanceof IOException)
					throw (IOException) thrown;	//rethrow it
				else if (thrown instanceof RuntimeException)
					throw (RuntimeException) thrown;	//rethrow it
				else {
					System.err.println("\nWrapper Exception:");
					e.printStackTrace();
					System.err.println("\nException wrapped up:");
					thrown.printStackTrace();
				}
			} catch (IllegalAccessException e) {
					e.printStackTrace();
			}
			if (IBM_DEBUG) System.err.println("IBM: me is "+
								(me==null?"null":"not null") );
			if (KELLY_DEBUG) {
				if (me instanceof java.awt.Container) {
					System.err.println("Got a java.awt.Container");
				}
			}
			if (DEBUG) System.err.println("Made class \"" +elementName+ "\"");
				//Constructor con = clazz.getConstructor(new Class[] {});	//no-arg con
				//con.setAccessible(true);
				//me	= con.newInstance(new Object[] {});	//no-arg again
		
	//6 Aug 2001 - need to store present alias *if* has readResolve.
			alias = putAlias(me);	//6 Aug 2001
			//putAlias(me);	//NOTE: do before any Strings in addAttr()
				//Be careful of this complexity!
				//To make consistent, may need to change the above
			meClassFrame = clazz; 
			
			if (superVersion) {	
				//meClassFrame = Class.forName(elementName);
				name = addPrim(addAttrTmp, me, false);	//there should be nothing.
				nameMissing = name.nameMissing;	//for setFinal at end
				ParserXML.Tag openSubclassTag = p.readTag();
				meClassFrame = Class.forName(ParseUtilities.descapeDollar(openSubclassTag.name));
				addAttrTmp = addName();	//should be subclass
				if (ADDATTR_NULL_DEBUG)
					System.err.println("addAttrTmp (in createObject)= "+addAttrTmp);
				nestedSubclass(addAttrTmp, meClassFrame, me);	//a test!!!
				String expectedName = ParseUtilities.escapeDollar(me.getClass().getName());
				ParserXML.Tag foundTag = p.readTag();	//expect to be leafClass
				checkCloseTag(expectedName, foundTag); 	//String, Tag
			}
		} else {	//a subclass - never happens, except when superVersion.
			meClassFrame = Class.forName(ParseUtilities.descapeDollar(elementName));
		}

    if (!superVersion) { 
			Vector localStackPrimStore = primStore;	//push old (see Explanatory memo)
			int localStackPrimStoreIndex = primStoreIndex; 
			primStore = new Vector(); //new context - NOTE: need for addAttr
			primStoreIndex = 0;
				//DESIGN DECISION: custom data Strings never in attributes!

			//Move up, along with addName, to before createObject
			//  ParserXML.Attr name = null;	

			//Expand and slice old "addAttr":
			//  name = addAttr(me, false);	//adds in all attr and gets NAME_TOKEN
			//Move getting the name part up to before instance is created:
			//  ParserXML.Attr[] addAttrTmp = addName();	//split it up
			name = addPrim(addAttrTmp, me, false);

	 		nameMissing = name.nameMissing;
			if (DEBUG && name!=null) System.err.println("Name: "+name);
			if (DEBUG) System.err.println("nameMissing: "+nameMissing);

		 /* Explanatory memorandum for custom data:
			*	--------------------------------------
			*	Custom data only occurs in classInstances (not null, alias, String,
			*	hash or vect).  It has two phases: (1). Store custom data in buffers
			*	(all attrs and contained, nested tags).  (2). Callback the objects's
			*	readObject() method, drawing from the buffers, via calling readObject
			*	within it readInt(), readLong() etc.  The buffer context needs to be
			*	stacked, because contained tags (field or custom data objects) may
			*	themselves contain custom data.  Both fields and custom data object are
			*	created the same; the difference is that custom data is stored in a
			*	stacked buffer.  This stack is implemented with the local var
			*	stackframe (CONSIDER: would an explicit stack be clearer?).
			*
			* In effect, each classinstance has two potential kinds of data - in
			* fields, and in buffers.  The buffer is therefore like a kind of series
			* of positional field, or typeless array.  This provides another view of
			* the need for stacking: each object has its own buffer, just as it has
			* its own field.
			*
			* Attributes are stored into their parent buffer when we are in their
			* "parent" object.  This is because attributes can never be parents -
			* there is no possibility of nesting (tho we still need to stack).  In
			* contrast, objects are stored into the parent buffer when we are in the
			* "child" object.  This is because we always change the context for
			* objects.
			*
			* This difference is caused by not changing context for attributes - it
			* might be good to regularise it by representing the present object
			* context as a "parent" when doing the attributes.
			*
			* Note: custom data String objects are never stored in attributes.  This
			* avoids the problem of determining that it is a String (and so needs to
			* create an alias, and be put in objStore, not primStore).  The type of
			* fields can be determined from the class; but not so for custom data.
			* Although the position does ultimately determine how it is read, this is
			* deferred (and so can cause problems for aliasing), and the attr/tag
			* split (seems) to not preserve the ordering - unless we know into which
			* to put the String beforehand (begging the question)!  Easier to simply
			* put it in a tag!!  Previous design comments:
			*
			* >This order is in fact lost already - it was lost when the Strings were
			* >included as Attr, instead of as objects... the simplest thing may well
			* >be to represetn *every* string as a wrapper... or, at least, to include
			* >every customized string as a wrapper...
			*
			* Note that we still need to stack attributes, even though they
			* themselves don't nest, since the object containing them will have
			* invocation of its readObject() method deferred until after any
			* contained tags (objects) are also processed, which need their own
			* "buffer field" for any custom data.
			*
			*/
	
			if (INTERNAL_DEBUG) {
				if (primStore!=null) {
					for (Enumeration keys=primStore.elements(); keys.hasMoreElements();) {
						Object a = keys.nextElement();
						System.err.println("In primStore: " +a+ " -- "
					 	+ a.getClass() + " -- " +a.hashCode());
					}
				}
			}

//CHILDREN TAGS; recursion, and a new context for buffering custom data
		 /*
			* A new custom data context is created, and then the children are done.
			* Each custom data child is buffered into this context - all siblings
			* into the same one.  This buffering is done in this routine
			* (createObject), with objStore.add().  Incidentally, they nest, so
			* memory usage is (aveObjectSize*depth).
			*/
			Vector localStackObjStore = objStore; //save locally, so recurse
			objStore = new Vector();	//this will get all the enclosed tags of "me"
			int localStackObjStoreIndex = objStoreIndex;	//also save it.
			objStoreIndex = 0;

			Object localMeStore = meStore; //Push onto program stack (frame)
			meStore = me;	//"pass arg", to defaultReadObject
			Class localMeClassFrameStore = meClassFrameStore; //Push
			meClassFrameStore = meClassFrame;
			boolean localEmptyTagStore = emptyTagStore; //Push
			emptyTagStore = name.emptyTag;
//addTags *is* the default!! -- now called from ObjIn!
//		if (!name.emptyTag) {	//ie if has elements (not <xxxx/>)
//			addTags(me);	//CENTRAL RECURSIVE dispatch: me am become the parent.
//		}

//CALLBACK object's readObject()
		 /*
			* The defaultReadObject() problem
			* -------------------------------
			* All contained tags are read; now invoke readObject() to get them.
			* Strictly, any fields should only be set if defaultReadObject() is
			* called.  But JSX doesn't work like this.  JSX sets them all if they are
			* present in the XML, before readObject() is even called.
			*
			* This rarely seems to be a problem - it would only be so if a class
			* *depended* on the recorded values not being set.  Usually, such a class
			* wouldn't write them in the first place - that is, it wouldn't call
			* defaultWriteObject() in writeObject(), and so the fields wouldn't be
			* present.  However, it might be relied upon in evolution of classes.
			*
			* A solution to this would be to not set the fields until after
			* readObject() is invoked; similar to as is presently done for the
			* custom data.  This means processing all fields twice - once to store,
			* and once to set into field.  Although it might also seem to create
			* a complete DOM tree replica of the object graph, it is not all in memory
			* at once, but only the path to the present object (on the stack).
			* (This does trace out an entire DOM tree, just not all at once).

			* Efficiency ("Don't" and "Not yet") - needn't do this when:
			*  - the parent is an Arrays, Vect or Hash.
			*  - there is no readObject() method (and so all fields are read)

			* The custom data Strings as attributes problem
			* ---------------------------------------------
			* There is a separate problem for custom data Strings in attributes:
			* There is no markup to indicate the type of attributes, so that we
			* don't know whether they are Strings or primitives, and therefore
			* don't know what to do with them.  For fields, we can derive the type
			* via Java reflection, based on the name.  But for custom data, we
			* don't know until the readObject() requests it.  However, for this to
			* work, we would need to have the custom data all in one buffer, rather
			* than split in two.
			*
			* [The difficulty is to do with splitting it into object/primitive, which
			* loses the relative ordering between them, before we can decide whether
			* it is an object or primitive!  Catch 22.  Begging the question.  This
			* information is in fact lost in the serialization into primitive vs
			* object - that's the split, not the buffering.  I'm a still a bit
			* confused about *exactly* what is wrong with this, which is why it's
			* not quite expressed right.  However, I'm convinced that it is an
			* insurmountable problem, due to the attr/object split.  The simple
			* resolution is just to always put custom data Strings into tags.]
			*
			* A solution to the buffering problem is to just do all the custom data
			* in one section, and so not need to buffer.  The readObject() could read
			* this directly, skipping named data.  This would require putting the
			* custom data at the end (so no buffering needed), or putting it into a
			* new "<custom>" tag - which would also need to go at the end. However,
			* the attr/tag split problem for Strings still remains.  This could be
			* solved by putting all custom data into tags - though we would need
			* someway to distinguish between primitives and their object wrappers.
			* Ie between Byte and byte.  Of course, we simply introduce <byte
			* valueOf=""/> tags and so on.

			* NOTE: in practice, a class not calling defaultReadObject() also should
			* not call defaultWriteObject().  If so, the fields are *not* serialized,
			* and so the defaultReadObject() has no effect, in practice.  This is not
			* the proper solution, but it may well work in all cases in practice.
			* Something that *does* need to be fixed for correct implementation of
			* the Java Serialization API.

			* SUMMARY  (rough, first draft)
			* -------
			* (1). Buffering of fields (as well as custom data) is *not* too memory
			* intensive - just a little stack overhead.  It will be *slightly* less
			* efficient.  It will solve the defaultReadObject() problem (though which
			* should never occur in practice, anyway).  NOT PLANNING TO DO YET.
			* (2). The attr/tag custom data String problem needs a dramatic change
			* to fix properly.  Since there seems no disadvantage to simply
			* serializing in a tag, that is what has been done.
			* (3). the custom data represents the positional fields of each object.
			* (4). fields are created and set as they are read.
			* (5). custom data is fully created and stored as it is read - then
			* processed after *all* tags have been read.
			* (6). tags (objects) are stored into a buffer belonging to the parent
			* object.
			* (7). attr (primitives) are stored into a buffer belonging to the
			* present object.
			*/

//CALLBACK OF readObject() for CUSTOM DATA
			boolean localDefaultCalledStore = defaultCalledStore; //Push
			defaultCalledStore = false;
			Method m;
			if (me instanceof Externalizable) {
				//try {
					((Externalizable)me).readExternal(oisSubclass);
				//} catch (IOException e) {
					//should propagate, but awkward in present architecture
				//} catch (ClassNotFoundException e) {
				//}
			} else {// regular treatment
					Vector classes = new Vector();	//NB: Mirror order - reverse.
					Class[] superClasses=XMLSerialize.getReversedSuperClasses(me.getClass());
					if (SUPER_DEBUG)
						System.err.println(me.getClass().getName()+" "+(!nameMissing?name.value:"<unnamed>")+ ": superClasses.length = "+ superClasses.length);
					for (int j=0; j<superClasses.length; j++) {  //traverse root to leaf
						Class c = superClasses[j];
						if ((m=XMLSerialize.getDeclaredMethod(c, "readObject", OIS_ARGS,
								Modifier.PRIVATE, Modifier.STATIC)) != null) {	//NB: ?clazz->c
							if (INTERNAL_DEBUG) System.err.print("\t--JSX deserialize--");
							try {
								m.invoke(me, readObjectArglist);	//might as well do them all!
								if (INTERNAL_DEBUG)
									System.err.println("\tinvoke "+m+" on "+me+" of " +me.getClass());
							} catch (InvocationTargetException e) {
								if (DEBUG) {
									System.err.println("JSX InvocationTargetException:");
									System.err.println("Object which is a: " +me.getClass());
									e.printStackTrace();	//should we just rethrow this?  YES!
								}
								//we just copied Java's ObjectInputStream handling of it!
								//It does all this in an "invokeObjectReader" method
								Throwable t = e.getTargetException();
								if (t instanceof ClassNotFoundException)
									throw (ClassNotFoundException)t;
								else if (t instanceof IOException)
									throw (IOException)t;
								else if (t instanceof RuntimeException)
									throw (RuntimeException) t;
								else if (t instanceof Error)
									throw (Error) t;
								else
									throw new Error("interal error");
							} catch (IllegalAccessException e) {
								defaultReadObject();	//just as if it were not there.
							}
						} else {
							defaultReadObject(); //call if no readObject() declared
								//potential BUG: will JSX work if the readObject doesn't call
								//defaultReadObject?  Doesn't there need to be some check, so
								//that it does get called, if it hasn't been already?
								//This is like the other side of ensuring it is only called
								//once.
						}
					}
					if (superClasses.length==0)	//ie for java.lang.Object
						defaultReadObject();	//a more elegant placement?
			}//end: regular (non-Externalizable) treatment
				//BTW: "Externalizable extends Serializable"

				//how is the best way (most modular) for readResolve to operate?
				//Can we move readResolve out of this nested position?
			if (!subclass) {
				if (NOV_DEBUG) System.err.println("readResolve is called!!!!!!!!!!!");
				me = readResolve(me, alias);	//defaults to me, if no readResolve
			}

			defaultCalledStore = localDefaultCalledStore; //pop
			//for defaultReadObject
			meStore = localMeStore; //pop
			meClassFrameStore = localMeClassFrameStore; //pop
			emptyTagStore = localEmptyTagStore; //pop

			//for Objects (tags)
			objStoreIndex = localStackObjStoreIndex;	//restore
			objStore = localStackObjStore;//restore objects for previous level

			//for Primitives (attr)
			primStoreIndex = localStackPrimStoreIndex;	//restore
			primStore = localStackPrimStore;//restore primitives for previous level
		}

		if (superVersion) {	//duplicate of above
			if (!subclass) {
				if (NOV_DEBUG) System.err.println("readResolve is called!!!!!!!!!!!");
				me = readResolve(me, alias);	//defaults to me, if no readResolve
			}
		}


//PLUG: CUSTOM DATA->BUFFER; FIELD->PARENT
//There is no real reason that this cannot be at the end.  It is even an
//appropriate place for it to be - just before it is returned, which is
//the alternate mechanism.
//--Note that this section uses variables that are set from above--
	//6 Aug 2001 - QUESTION: can we move both set and add to after readObject?
	//They both seem to only relate to the parent, which is invariant
	//across the readObject invocation - great if could confirm this.
	//DEPENDS ON: me, parent, nameMissing, nameValue, a nested objStore.
		//addTags is done by the call to defaultReadObject(), which the object's
		//readObject() needs to invoke to setup the custom data for reading.
		//Basically: the "parent" context (parent, objStore) is set up around
		//the readObject/defaultReadObject calls... ah! The naming shows that it is
		//the central thing - even when not explicitly called, it still is called.
		//Our confusion is mainly caused by addAttr, which is because we mix
		//info *about* object (name, class, etc) with *content* of the object 
		//itself - ie primitives and strings.  This is conceptually inconsistent.
		 if (!subclass) { //ie if not ob-name="subclass"
			if (parent!=null) { //can't plug if no parent (vect, hash, array do sep)
				//(1). BUFFER (CUSTOM DATA)
	 			if (nameMissing) {//no name
					if (parent.getClass()!=Vector.class &&			//not vector
							parent.getClass()!=Hashtable.class &&		//not hashtable
							!parent.getClass().isArray()) {						//not array
						objStore.add(me);	//store custom *Object* in buffer
						if (INTERNAL_DEBUG)
							System.err.println("custom obj["+objStoreIndex+"],"+objStore);
					}
				}
				//(2). INTO CLASS AS FIELD
				else {
					if (DEBUG_CUSTOM) {
						System.err.println("getting field '"+name.value+"'");
					}
					//try {
						currentGetField.putObj(name.value, me);	//prim/obj?
						Field f = getAllField(parent.getClass(), name.value);
						setFinal(f, parent, me);
						//f.set(parent, me);
					// catch (NoSuchFieldException e) 
					//
				}
			}
		 } //subclass

			//System.err.println("me being returned: "+me);
			//try { throw new Exception(); } catch (Exception e) {e.printStackTrace();}
			
			return me;
		}
	}

/**-----------------------------------------------------------------------
	* Convenience methods for the common tasks of checking for open or close
	* tags.  Note that these operate by throwing exceptions... in effect, we
	* have factored out the exception generating code... ;-)
	**/
  private void checkCloseTag(String expectedName, ParserXML.Tag tag) throws ParserXML.ExceptionXML {
		if (tag.start)	//we want close - this is not my beautiful closet!
			throw new ParserXML.ExceptionXML("Wanted a close tag, got an open tag", "</"+expectedName+">","<"+tag.name+">");
		else if (!tag.name.equals(expectedName))
				throw new ParserXML.ExceptionXML("</"+expectedName+">","</"+tag.name+">");
	}
  private void checkOpenTag(String expectedName, ParserXML.Tag tag) throws ParserXML.ExceptionXML {
		if (!tag.start)	//we want start - this is not my beautiful start!
			throw new ParserXML.ExceptionXML("Wanted an open tag, got a close tag", "</"+expectedName+">","<"+tag.name+">");
		else if (!tag.name.equals(expectedName))
				throw new ParserXML.ExceptionXML("</"+expectedName+">","</"+tag.name+">");
	}



/** ---------------------------------------------------------------------------
	* nest
	* ----
	* If it returns to here, means either: open sub-class, or a close tag.
	* Question: who handles the non-subclass close tags?
	*
	*	(Consider: for convenience, we could provide both Class and String.)
	**/

	private void nestedSubclass(ParserXML.Attr[] addAttrTmp, Class classFrame, Object me) throws IOException, ClassNotFoundException {
		//it should *only* be here if a classinstance
		if (NESTED_DEBUG)
			System.err.println(">>>> entering nested");

			//meClassFrameStore = classFrame;
			if (ADDATTR_NULL_DEBUG)
				System.err.println("addAttrTmp = "+addAttrTmp);
		ParserXML.Tag tag = readObject(addAttrTmp, me, classFrame);	//attr & tags...
		addAttrTmp = defaultReadObject_attrs_out;	//pass it out

		//DOWN (nest deeper)
		if (tag.start) {	//open
			if (NESTED_DEBUG)
				System.err.println("nestedSubclass got an open subclass tag: "+tag);
				//assume "sub-class": the only open tag defaultReadObject returns on
			Class nextClass = Class.forName(ParseUtilities.descapeDollar(tag.name));
			if (ADDATTR_NULL_DEBUG)
				System.err.println("addAttrTmp (DOWN)= "+addAttrTmp);
			nestedSubclass(addAttrTmp, nextClass, me);	//attr & tags...
			ParserXML.Tag endTag = p.readTag();	//when return, must be a close tag
			//it should be a close tag, matching the open tag
			if (endTag.start)	//if not a close tag
				throw new ParserXML.ExceptionXML(
					"</"+classFrame.getName()+">","<"+endTag.name+">");
			else
				checkCloseTag(ParseUtilities.escapeDollar(classFrame.getName()),endTag);
 				//if (!classFrame.getName().equals(endTag.name))	//end, but wrong one
					//throw new ParserXML.ExceptionXML(
						//"</"+classFrame.getName()+">","</"+endTag.name+">");
			if (NESTED_DEBUG)
				System.err.println("tag when return from nestedSubclass: "+tag);
		} else {	
		//UP (return)
			if (NESTED_DEBUG)
				System.err.println("nestedSubclass got a close tag: "+tag);
			String expectedTag = ParseUtilities.escapeDollar(classFrame.getName());
			if (!tag.name.equals(expectedTag))
				throw new ParserXML.ExceptionXML(
					"</"+expectedTag+">","</"+tag.name+">");
		}
		if (NESTED_DEBUG)
			System.err.println("<<<< leaving nested, with tag: "+tag);
	}


//A convenience wrapper for argument passing.
		Vector defROStack = new Vector();
		Vector defROCalledStack = new Vector();
	ParserXML.Tag readObject(ParserXML.Attr[] addAttrTmp, Object me, Class meClassFrame) throws IOException, ClassNotFoundException {
		if (ADDATTR_NULL_DEBUG)
			System.err.println("addAttrTmp =" +addAttrTmp);

		//(1). set up call to defaultReadObject, from whichever path.
			//Note that these are always direct, so don't need to stack.
		defaultReadObject_attrs_in = addAttrTmp;	//pass arg in
		defaultReadObject_me = me;	//pass arg in
		defROStack.add(null);	//Make room on top of stack - in case no defRO call
		defROCalledStack.add(null);	//null means "not called"
		// defaultReadObject_ver2();	//*always* called!
			//condition this call, so only call directly, or when no RO
		// ParserXML.Tag entryTag = defaultReadObject_tag;
		Class clazz = meClassFrame;
			//me.getClass();

		Method readObjectMethod = XMLSerialize.getDeclaredMethod(clazz,
"readObject", OIS_ARGS, Modifier.PRIVATE, Modifier.STATIC);
		if (readObjectMethod==null) {
			defaultReadObject_ver2();	//called by default (if no readObject)
		} else {
			if (NOV_DEBUG) System.err.println("using correct class frame now (I believe)");
			if (NESTEDREADOBJECT_DEBUG) System.err.print("\t--JSX deserialize--");
			try {
				readObjectMethod.invoke(me, readObjectArglist);
				if (NESTEDREADOBJECT_DEBUG)
					System.err.println("\tinvoke "+readObjectMethod+" on "+me+" of " +me.getClass());
			} catch (InvocationTargetException e) {
				if (DEBUG) {
					System.err.println("JSX InvocationTargetException:");
					System.err.println("Object which is a: " +me.getClass());
					e.printStackTrace();	//should we just rethrow this?  YES!
				}
				//we just copied Java's ObjectInputStream handling of it!
				//It does all this in an "invokeObjectReader" method
				Throwable t = e.getTargetException();
				if (t instanceof ClassNotFoundException)
					throw (ClassNotFoundException) t;
				else if (t instanceof IOException)
					throw (IOException)t;
				else if (t instanceof RuntimeException)
					throw (RuntimeException) t;
				else if (t instanceof Error)
					throw (Error) t;
				else
					throw new Error("internal error");
			} catch (IllegalAccessException e) {
				System.err.println("IllegalAccessExcetion - please report this error");
				e.printStackTrace();
				//defaultReadObject();	//can this happen?
			}
			//POP: at end of RO, pop stack, to get the most recently invoked defRO
			//tag = defRO_tagStack.remove(defRO_tagStack.size()-1);
		}
		ParserXML.Tag entryTag;
		entryTag = (ParserXML.Tag) defROStack.get(defROStack.size()-1);
		//(1). If entryTag==null, it means that defRO was *not* called...
		//(2). If defRO was not called, it means that we need to eat the def data,
		// or just ignore it (in fact, there shouldn't be any def data)
		//(3). We call "eatDef()", which leaves us in the same state as 
		//the real def.  In fact, as a short-cut fake, we could just call defRO...
		//(4). What we *should* do is copy defaultReadObject_ver2, and 
		//remove the forwarding bits... not sure how to do this... just set
		//"obj" to null?
		if (entryTag==null) {
			defaultReadObject_ver2();	//if not called.  This can happen if:
				//(1). RO
				//(2). not called explicitly within RO
				//(3). no writeObject or writeInt etc within RO
				//-> we need to call it now, to set up the next part
				//set flag for writeObject (ie deserialize) and writePrim
			entryTag = (ParserXML.Tag) defROStack.get(defROStack.size()-1);
		}



		if (NOV_DEBUG) System.err.println("------ entryTag just got from defROStack: "+entryTag);

		//ParserXML.Tag entryTag = defaultReadObject_tag;	//assuming that def
			//was called... if not, what is the value of this?

		//eat any opt tags that appear (eg writeObject made them, but no readObject)
			if (KELLYSUPER_DEBUG)
				System.err.println("called readObject, entryTag is "+entryTag);
			if (entryTag.name.equals(XMLSerialize.OPT_DATA_TOKEN)) {
				ParserXML.Tag tag = p.readTag();
				if (tag.start && tag.name.equals(XMLSerialize.OPT_DATA_TOKEN)) {//eat <opt>
					addAttr(null, false);	//eat it
					tag = p.readTag();	//read next one: in case <opt></opt>
				}
				checkCloseTag(XMLSerialize.OPT_DATA_TOKEN, tag); 	//String, Tag
				//defaultReadObject_tag = p.readTag();
				defROStack.set(defROStack.size()-1, p.readTag());	//replace the top null
				if (((ParserXML.Tag)defROStack.get(defROStack.size()-1)).start)
					defaultReadObject_attrs_out = addName();	//to set up...
			}
		//What if there is a writeObject, but no readObject?
			if (KELLYSUPER_DEBUG)
				System.err.println("entryTag is "+entryTag);

//		if (readObjectMethod==null) {	//separate, not "else", for clarity
//			defaultReadObject_ver2();
//		}
		//return defaultReadObject_tag; //get return value out
		return (ParserXML.Tag)defROStack.remove(defROStack.size()-1);	//finally pop
	}


/** ------------------------------------------------------------------------
 	* defaultReadObject
	* -----------------
	* For superVersion, will be called directly from object's readObject()...
	* Switch, according to which version, for back-compatibility.
	**/
	void defaultReadObject() throws IOException, ClassNotFoundException {
		if (superVersion) {
			;	//disable, so that always called.
			defaultReadObject_ver2();	//new, nesting one
				//re-enable this, so that the *time* of call is correct
		} else {
			defaultReadObject_ver1();	//old
		}
	}


/** new Version of defaultCalledStore
	* Should it still have "defaultCalledStore" and "emptyTagStore"?
	*
	**/
	//boolean defaultCalledStore;	//so only called once per object  used??
	Object defaultReadObject_me;	//arg in
	//Class defaultReadObject_classFrame; //arg in
	//ParserXML.Tag defaultReadObject_tag; //return value
	ParserXML.Attr[] defaultReadObject_attrs_in; //arg
	ParserXML.Attr[] defaultReadObject_attrs_out; //return value
	//boolean emptyTagStore; //?
//Called from ObjIn and also from here
	void defaultReadObject_ver2() throws IOException, ClassNotFoundException {
		defROCalledStack.set(defROCalledStack.size()-1, Boolean.TRUE);
			//just to say that this work has been done
		Object me = defaultReadObject_me;	//"passing" these two arguments.
		ParserXML.Attr[] addAttrTmp = defaultReadObject_attrs_in;
		if (ADDATTR_NULL_DEBUG)
			System.err.println("addAttrTmp = "+addAttrTmp);
		addPrim(addAttrTmp, me, false);

		while (true) {
			ParserXML.Tag tag = p.readTag();
			if (tag==null) throw new EOFException("Premature EOF");
			if (tag.start) {
				if (tag.name.equals(XMLSerialize.OPT_DATA_TOKEN)) {
					if (NOV_DEBUG) System.err.println("got a opt data tag");
					addPrim(addName(), me, false);	//just clear end of opt tag
						//add a "discard" flag arg?
					//defaultReadObject_tag = tag;	//needed!
					defROStack.set(defROStack.size()-1, tag);	//replace the top null
							//PUSH tag for readObject
					return;
				} else {
					addAttrTmp = addName();//to detect "sub-class"...
					ParserXML.Attr name = addAttrTmp[1];
					if (name!=null && name.value.equals(XMLSerialize.SUPER_TOKEN)) {
						if (NOV_DEBUG) System.err.println("defRO got an open subclass tag: "+tag);
						defaultReadObject_attrs_out = addAttrTmp;
						//defaultReadObject_tag = tag;
						defROStack.set(defROStack.size()-1, tag);	//replace the top null
							//PUSH tag for readObject
						return;	//break loop
					} else {
						//or falls through to here:
						if (NOV_DEBUG) System.err.println("defRO got an open non-subclass tag: "+tag);
						createObject(tag, addAttrTmp, me);	//need classFrame eventually
						//add a flag arg to createObject - "discard", to eatDef
					}
				}
			} else { //ie close tag ---- no addAttrTmp is appropriate!
				if (NOV_DEBUG) System.err.println("defRO got a close tag: "+tag);
				//defaultReadObject_tag = tag;
				defROStack.set(defROStack.size()-1, tag);	//replace the top null
							//PUSH tag for readObject
							//defRO_tagStack.add(tag);
				return;  //break loop
			}
		} //while
 	//Never gets here! - infinite while, only exits by return or exception
	}
/* a confusion:
	tag.name - changes with the type
	attr.value - changes with the name
*/


	//these were static, and may have made JSX un-thread-safe for 
	//Jean-Chrsitophe Salome, from Paris, 10 Oct 2001 (JSX's 1st birthday)
	boolean defaultCalledStore;	//so only called once per object
	Object meStore; //pass arg to object's readObject()
	Class meClassFrameStore;
	boolean emptyTagStore;
//Called from ObjIn and also from here
	void defaultReadObject_ver1() throws
			IOException,
			//NoSuchFieldException,
			ClassNotFoundException {
		//NotActiveException 
		if (!defaultCalledStore) {	//only do it once, until reset
			 defaultCalledStore = true;
			if (!emptyTagStore) {
					//addTags(meStore, meClassFrameStore); //<--- before absorbed
				Object me = meStore;	//"passing" these two arguments.
				Class meClassFrame = meClassFrameStore;
				ParserXML.Tag tag;
				while ( ((tag=p.readTag())!=null) && (tag.start))	{
					createObject(tag, me);
				}
				if (tag==null) throw new EOFException("Premature EOF");
				else {
					String className = meClassFrame.getName();	//set for class create
					String rightTag = ParseUtilities.escapeDollar(className);
					if (!tag.name.equals(rightTag))
						throw new ParserXML.ExceptionXML(
							"</"+rightTag+">","</"+tag.name+">");
				}
			}
		}
	}


	/**--------------------------------------------------------------------------
		* addTags <------- absorbed into defaultReadObject() above: simpler
		* -------
		* Handles just tags (no attrs)
		* CONSIDER: is there an elegant way to reuse this code in Hashtable?
		* This is where the end tag is read
		* Perhaps pass in the open tag explicity, rather than (inefficiently)
		* recalling it from the object itself.  Why?  Because this doesn't work
		* for null objects.
		**/
/*
	private void addTags(Object me, Class meClassFrame) throws
			ClassNotFoundException,
			//NoSuchFieldException,
			IOException {
		ParserXML.Tag tag;
		while ( ((tag=p.readTag())!=null) && (tag.start))	
			//createObject(tag.name, me);	//add this new tag to me
			//createObject(tag, me, meClassFrame);	//add this new tag to me //30Apr2001
			createObject(tag, me);	//add this new tag to me //30Apr2001
//May10: NOTE!!!  We throw away the returned value, because it is already
				//plugged in for us! (done like this, because parent not known here)
				//Returned value is only used by: array, vector, hashtable
		if (tag==null)
			throw new EOFException("Premature EOF");
//		if (me==null) {		//dangling else!
//			if (!tag.name.equals(XMLSerialize.NULL_TOKEN))	//special case of null
//				throw new ParserXML.ExceptionXML(
//					"</"+XMLSerialize.NULL_TOKEN+">","</"+tag.name+">");
//		}
		else {
			String className = meClassFrame.getName();	//set for class create
			String rightTag = ParseUtilities.escapeDollar(className);
			//if (!tag.name.equals(ParseUtilities.escapeDollar(me.getClass().getName())))
			if (!tag.name.equals(rightTag))
				throw new ParserXML.ExceptionXML(
					"</"+rightTag+">","</"+tag.name+">");
		}
	}
*/



/** ---------------------------------------------------------------------------
	* readResolve
	* -----------
	* Reflectively invoke the present objects's readResolve().
	* Return value can be: an object or an exception.  We are using an 
	* Exception to flag the result, instead of trying to encode it in the
	* return value itself (by wrapping it, for example - which is sort of how
	* Exceptions work, anyway.
	* QUESTION: If we checked for it separately, we can choose whether to cache
	* or not...
	* "parameterless" = no-arg
	**/
	Object readResolve(Object me, String alias) throws ObjectStreamException {
		if (READRESOLVE_DEBUG)
			System.err.println("entered readResolve, with "+me+" and alias="+alias);
		try {
			Method readResolveMethod = null;
			try {
				readResolveMethod = me.getClass().getDeclaredMethod("readResolve", new Class[]{});
			} catch (NoSuchMethodException e) {
				//Not an error - just means that it doesn't implement readResolve
				if (READRESOLVE_DEBUG)
					System.err.println("No readResolve() found - checking super");
				try {
					Class superClass = me.getClass().getSuperclass();
					if (superClass!=null)
						readResolveMethod = superClass.getDeclaredMethod("readResolve", new Class[]{});
				} catch (NoSuchMethodException e2) {	//of super
					return me; 	//forget it - invoke nothing
				}
				if (!checkSuperMethodAccess(readResolveMethod, me.getClass()))
					return me; //forget it - invoke nothing
			}//end of super - method is now set correctly
			readResolveMethod.setAccessible(true);
			Object newme = readResolveMethod.invoke(me, new Object[]{}); //no arg
			updateAlias(newme, alias);	//only need to update if it really is new.
			if (READRESOLVE_DEBUG)
				System.err.println("readResolve returned "+newme);
			return newme;	//if no readResolve - the object is simply unchanged!

		} catch (InvocationTargetException e) {
			System.err.println("JSX InvocationTargetException:");
			//e.printStackTrace();	//not sure how this should be handled...
			//is this the best way to handle it?  What does Serialization do?
			//Probably best to throw whatever is thrown.
			try {
				throw (ObjectStreamException) e.getTargetException(); //uncork it
				//contract of readResolve guarantees that it only throws this class
				//I should check that it *does* throw only this kind...
			} catch (ClassCastException f) {
				throw new Error("readResolve() of "+me.getClass()+ " threw an '"+e.getTargetException()+"'. It should only be of type ObjectStreamException");
			}
		} catch (Exception e) {
			System.err.println("not an InvocationTargetException");
			e.printStackTrace();
		}
		return null;	//to signal some error.
	}

//Adapted from ObjectInputStream
	static private boolean checkSuperMethodAccess(Method scMethod, Class ofClass) {
		if (scMethod == null) {
			return false;
		}
		int supermods =  scMethod.getModifiers();
		if (Modifier.isPublic(supermods) || Modifier.isProtected(supermods)) {
			return true;
		} else if (Modifier.isPrivate(supermods)) {
			return false;
		} else {
		// check package-private access.
			return isSameClassPackage(scMethod.getDeclaringClass(), ofClass);
			//NB: ofClass is the *runtime* (lowest, leaf) class, *not* the super
			//return false;	//cheat for now - inherited default visible readResolve
				//won't work properly at the moment.
		}
	}
    /* Will not work for array classes. */
	static private boolean isSameClassPackage(Class cl1, Class cl2) {
		if (cl1.getClassLoader() != cl2.getClassLoader()) {
			return false;
		} else {
			String clName1 = cl1.getName();
			String clName2 = cl2.getName();
			int idx1 = clName1.lastIndexOf('.');
			int idx2 = clName2.lastIndexOf('.');
			if (idx1 == -1 || idx2 == -1) {
				/* One of the two doesn't have a package. Only return true
				* if the other one also does not have a package.
				*/
				return idx1 == idx2;
			} else {
				return clName1.regionMatches(false, 0, 
				clName2, 0, idx1 - 1);
			}
		}
	}







	/**--------------------------------------------------------------------------
		* deserializeNull()
		* -----------------
		* Called from createObject(), to separate out the parsing code.
		**/
	//private NamedObject deserializeNull(Object parent) throws
	private NamedObject deserializeNull(ParserXML.Attr[] addAttrTmp, Object parent) throws
			ClassNotFoundException,
			//NoSuchFieldException,
			NotActiveException,
			IOException {
		String name = null;
		Object me = null;
		boolean named = false;
		//ParserXML.Attr firstAttr = addPrim(me, false);	//gets NAME_TOKEN
		ParserXML.Attr firstAttr = addPrim(addAttrTmp, me, false);	//gets NAME_TOKEN
		if (!firstAttr.nameMissing) { //see dup in createArray
			named = true;
			name = firstAttr.value; //NB: name of field is *value* of attr
		}
		if (!firstAttr.emptyTag) {	//if not <null/>
					//then start & end: <null></null> - read the end one.
			//addTags(null);
			ParserXML.Tag tag = p.readTag();
			if (tag==null)
				throw new EOFException("Premature EOF");
			else if (tag.start) //if start tag...
					throw new ParserXML.ExceptionXML(
						"</"+XMLSerialize.NULL_TOKEN+">","<"+tag.name+">");
			else //if an end tag... it needs to match the <null> start tag...
				if (!tag.name.equals(XMLSerialize.NULL_TOKEN)) {	//special case of null
					throw new ParserXML.ExceptionXML(
						"</"+XMLSerialize.NULL_TOKEN+">","</"+tag.name+">");
				}
			}

	//again, code dup from createArray
		if (parent!=null) {
			if (named) {
				if (name==null) {	//should never happen
					throw new IOException("name: No " +XMLSerialize.NAME_TOKEN+ " attribute found for "+parent.getClass());
				}
				//try {
					currentGetField.putObj(name, me);	//prim/obj?
					Field f = getAllField(parent.getClass(), name);
					setFinal(f, parent, me);
					//f.set(parent, me);
				//} catch (NoSuchFieldException e) {
				//}
 				return new NamedObject(name, me, named);
			} else { //!named -> (internal: parent, but no name)
 				return new NamedObject(name, me, named);
			}
		} else {	//parent==null -> (root, array, hash, vect)
 			return new NamedObject(name, me, named);
		}
	}




	/**--------------------------------------------------------------------------t
		* getAllField()
		* -----------
		* Named in accord with the JSX serialization method "getAllFields()".
		* Returns private and superclass fields.

		* If NoSuchFieldException, we call NoSuchFieldHander... and re-try just
		* *once* again:
		*	If NoSuchFieldException,
		*	  if handler: get the new field and try once more.
		*	    if another NoSuchFieldException, then fail noisily.
		*	  else (ie no-handler): fail silently.
		*	Thus, adding the handler is switches on debug exceptions (ie non-silent)
		*
		**/
	private	Field getAllField(Class c, String fieldName)
		throws SecurityException {
		Field field = null;
		try {
			field = attemptGetAllField(c, fieldName);	//try once...
		} catch (NoSuchFieldException firstE) {//if field not there, try handler...
			if (cfg!=null && cfg.noSuchFieldHandler!=null) {//...provided there is one...
				if (CFG_DEBUG) System.err.println("cfg = "+cfg);
				String mappedFieldName = null;
				try {
					mappedFieldName = cfg.noSuchFieldHandler.noSuchField(c, fieldName);
				} catch (NoSuchFieldException secondE) {
					throw new IllegalArgumentException(secondE+"");	//wrap
				}	//end: lookup field
				try {
					field = attemptGetAllField(c, mappedFieldName); //try twice...
				} catch (NoSuchFieldException secondE) {
						//if handler cannot handle it: "Noisy fail":
					throw new IllegalArgumentException(
						"Mapped '"+fieldName+"'" +
						" to '"+mappedFieldName+"'" +
						", but neither found in "+c+": "+secondE);	//poor man's exception wrapping
				}	//end: third NoSuchFieldException
			}
			//if no handler, then: "Silent fail".
		}	//end: first NoSuchFieldException
		return field;
	}


	/**--------------------------------------------------------------------------t
		* attemptGetAllField()
		* --------------------
		* Named in accord with the JSX serialization method "getAllFields()".
		* returns private and superclass fields.
		* Called by "getAllFields" above.
		*
		* DANGER: it could find a field from a superclass that customises its
		* own serialization.  This shouldn't happen. It really should check for
		* the readObject() method, and if present, disregard that declared class.
		*
		* We need to parse any class-qualified fields, and we can get these
		* directly from the named class - some unexpected efficiency.
		*
		* NOTE: should not consider any fields that are static and/or transient...
		*
		*/
	private	Field attemptGetAllField(Class c, String fieldName) throws
			NoSuchFieldException,
			SecurityException {
		if (fieldName==null)
			throw new NoSuchFieldException("field name had value 'null'");
		Field f = null;
		if ((f=getShadowedField(fieldName))!=null) {	//if class-qualified
			if ((f.getModifiers() & XMLSerialize.NOT_SERIALIZED) == 0) { //don't skip
				f.setAccessible(true);		//so can see private fields (Java 1.2)
				return f; //then just grab it
			}
			else throw new NoSuchFieldException(f +": this field evolved to non-serializeable, it seems.");
		}
			//else continue on
		Class initc = c;	//store it for error
		do {
			try {
				f = c.getDeclaredField(fieldName);
			} catch (NoSuchFieldException e) {	//do nothing; it's OK
			} catch (Exception e) {
				System.err.println(e +": No such field '"+ fieldName +"' in "+ c +" (superclass of " +initc);
			}
		} while ( f==null && (c = c.getSuperclass()) != Object.class);
		if (f==null)
			throw new NoSuchFieldException("JSX can't find the field "
			+((fieldName==null)?"<null>":"'"+fieldName+"'")
			+ ".  Not in " +initc+ ", or private, or in superclasses\n"
			+ "Could it be that the class lacks a readObject() method?");
		if ((f.getModifiers() & XMLSerialize.NOT_SERIALIZED) == 0) { //don't skip
			f.setAccessible(true);		//so can see private fields (Java 1.2)
			return f; //then just grab it
		}
			else throw new NoSuchFieldException(f +": this field evolved to non-serializeable, it seems.");
	}


	static Field getShadowedField(String s) throws NoSuchFieldException {
		int dotpos = s.lastIndexOf('.');
		if (dotpos==-1) return null;	//signal: couldn't find (clarity?)
		String classQual = s.substring(0, dotpos);	//front
		String fieldName = s.substring(dotpos+1);	//end (skip the dot)
		Class clazz = null;
		try {
			clazz = Class.forName(ParseUtilities.descapeDollar(classQual));
		} catch (ClassNotFoundException e) {
			System.err.println("The class-qualification of this shadowed field was not found: "+ s);
			e.printStackTrace();
		}
		return clazz.getDeclaredField(fieldName);
	}


/**==========================================================================**/
/**======================= ARRAY SECTION ====================================**/
/**==========================================================================**/

	/**--------------------------------------------------------------------------
		* createArray
		* -----------
		* Create one Array object.
		* Need to parse the whole tag name to know the component type,
		* 	eg: <ArrayOf-ArrayOf-int ..../>
		* that is: (ArrayOf-)*Type
		* 	(actually use a constant ARRAY_TOKEN = "ArrayOf-")
		* The mapping is:
		* 	(ArrayOf-)*Type -> Type([])*

		* It counts the ARRAY_TOKENS, extracts the type, then:
		*  Array.new_Instance(Class.getClass(baseType), new int[count])
		* NB: we avoid the complex disambiguation constraint on tok, by name choice

		* Complexity/Issues:
		*	 - Array of objects (nested tags)
		*	 - Array of primitives (attrs)
		*  - named: a field (parent is fields of a classInstance)
		*  - parent: only if a field or internal
		*  - unnamed: root, hash, vect, or array, internal
		*  - wrinkle 1: if parent and unnamed, then internal
		*  - wrinkle 2: if primitive, name attr can be before or after length attr
	  * NB: Internals are read in before we *know* they are internal, and they
	  * can *only* occur in classInstances, so they will have a parent.

		* Refactor for loose-coupled symbolic bindings
		* --------
		* Move all returned values to explicitly named variables for clarity
		*  - less efficient, but Knuth's first law of optimization is "don't".
		**/
	//private NamedObject createArray(String tagName, Object parent) throws
	private NamedObject createArray(ParserXML.Attr[] addAttrTmp, String tagName, Object parent) throws
			ClassNotFoundException,
			//NoSuchFieldException,
			IOException {
		//Loose-coupling: move returned values to named variables
		int length = -1; //how many components of array
		String name = null;	//field name of array (if parent is a classInstance)
			//NB: Name may be in firstAttr or secondAttr, for arrays of primitives
		boolean named = false; //redundant clarity (named==(name!=null))
			//named==true: if a *field* (of a classInstance)
			//named==false: if *internal* info (of a classInstance),
			//		--or-- if *root* or *element* (of array, hash or vector).
		boolean emptyTag = false; //true=="<tag/>", false=="<tag>...</tag>"

		//ParserXML.Attr firstAttr = addAttr(null, true); //object=null, isArray=true
		ParserXML.Attr firstAttr = addPrim(addAttrTmp, null, true); //object=null, isArray=true
		length = firstAttr.length;
		if (!firstAttr.nameMissing) { //if present, get name
			named = true;
			name = firstAttr.value; //NB: name of field is *value* of attr
		}
		emptyTag = firstAttr.emptyTag;

		//If an array of primitives, call addAttr twice:
			//(1). addAttr, to get length (and name, if present at start)
			//(3). create array
			//(2). addAttr, to get elements/attributes (and name, if present)
			//	NB: need length to create array
			//	NB: don't read elements til array created (could buffer in a Vector)
			//	NB: store name til after elements (attributes) are read

			//THUS: in addAttr (for array of primitives):
			// (1). addAttr to return after seeing "length"; and
			// (2). to call addAttr a second time, treating attr as indices.

			//THUS: in addAttr (for array of objects):
			// (1). reads the whole tag, including the ">" at the end of the tag.
		int dim = 0;
		int index=0;
		String tok = XMLSerialize.ARRAY_TOKEN;
		int skip = tok.length();
		//(1). take off the leading "ArrayOf-"'s (by incrementing an index for eff)
		while (tagName.startsWith(tok, index)) { //could be > eff if assume tok qual
			dim++;														//and just check first char...
			index+=skip;
		}
		if (DEBUG) System.err.println("dim= " +dim);
		String type = tagName.substring(index);	//get the type
		if (DEBUG) System.err.println("type= " +type);
		Object me = null;
		//try {
			int[] dimsArray = new int[dim];
			dimsArray[0] = length;		//NOTE use of length parsed from attributes
			Class comp;
			//shockingly inefficient!  This is probably why they used the 'L'
			if (type.equals("int")) comp = int.class;
			else if (type.equals("double")) comp = double.class;
			else if (type.equals("boolean")) comp = boolean.class;
			else if (type.equals("char")) comp = char.class;
			else if (type.equals("byte")) comp = byte.class;
			else if (type.equals("short")) comp = short.class;
			else if (type.equals("long")) comp = long.class;
			else if (type.equals("float")) comp = float.class;
			else comp = Class.forName(ParseUtilities.descapeDollar(type));
			me = Array.newInstance(comp, dimsArray);
			putAlias(me);
		//} catch (Exception e) {
			//System.err.println("Illegal type as Array base component, it seems; "+e);
		//}
		if (DEBUG) System.err.println("type= " +me.getClass().getName());
		if (DEBUG) System.err.println("len= " +Array.getLength(me));


		ParserXML.Tag tag;
		ParserXML.Attr secondAttr = null;
		//If array of primitives...
		if (dim==1 && me.getClass().getComponentType().isPrimitive()) {
			//secondAttr = addAttr(me, true); //object=me, isArray=true
			addAttrTmp = addName();	//for *second* call
			secondAttr = addPrim(addAttrTmp, me, true); //object=me, isArray=true
				//addAttr now reads primitive attrs into array
			length = -1;	//shouldn't be used again (just did it in addAttr)
			if (!secondAttr.nameMissing) { //was name mixed in with attrs?
				if (named) throw new IOException("two names: "
					+"'"+firstAttr.value+"' and '"+ secondAttr.value+"' "
					+"in "+parent.getClass());
				named = true;
				name = secondAttr.value;	//NB: name of field is *value* of attr
			}
			emptyTag = secondAttr.emptyTag;
			//(1). above allows "obj-name" to be before or after elements.
			//(2). "length" must always preceed the elements (alt: vector buffer)
			//(3). firstAttr/secondAttr seems only to be needed for emptyTag.
			if (DEBUG) System.err.println("after setting primitive array, saw: " + secondAttr);
		}
 		//If array of objects...
		else {
			//do the contained objects (as does Vector) - it dispatches in a loop
			//NOTE  the recursion runs through create object, which tests isArray,
				//so as to automatically bottom out. AND WHAT A HACK IT IS!
			ParserXML.Attr end = p.readAttr(); //slurp up the left-over '>' after length read.
			if (DEBUG) System.err.println("after starting reference array, saw: " + end);
				//NB: this implies that the name (if present) *must* precede length
			for (int i=0; i<length; i++) {
				//System.err.println("i = "+i);
				Object obj;
		 		tag=p.readTag();
				if (tag==null) throw new EOFException("Premature EOF");
				if (!tag.start)
					throw new IOException("Expected "+length+" elements in array; read only"+i);
				if (tag.name.equals("java.lang.String")	//bad: doubles up
						|| tag.name.equals("java.lang.Integer")
						|| tag.name.equals("java.lang.Double")
						|| tag.name.equals("java.lang.Boolean")
						|| tag.name.equals("java.lang.Character")
						|| tag.name.equals("java.lang.Byte")
						|| tag.name.equals("java.lang.Short")
						|| tag.name.equals("java.lang.Long")
						|| tag.name.equals("java.lang.Float")
					)
					obj = deserializeWrapper(tag, null).object;
				else
					//obj = createObject(tag.name, null);	//null=orphan object
					obj = createObject(tag, null);	//null=orphan object //30Apr2001
				if (DEBUG) System.err.println("Adding {"+obj+"} to Array");
				Array.set(me, i, obj);
			}	//for loop
		}	//tags (for refernce base type - array or class)

		if (!emptyTag) {		//if not <..../>, check for close tag
			tag=p.readTag();
			if (tag==null) throw new EOFException("Premature EOF");
			if (tag.start)
				throw new IOException("End of array");
			if (!tag.name.equals(tagName))	//check for matching close tag
				throw new ParserXML.ExceptionXML(
					"</"+tagName+">","</"+tag.name+">");
		}
//parent is null (this is an "orphan") for:
	//root, arrays, hash and vect
	//NB: Internals are read in before we *know* they are internal, and they
	//can *only* occur in classInstances, so they will have a parent.

		if (parent!=null) {
			if (named) {
				if (name==null) {	//should never happen
					throw new IOException("name: No " +XMLSerialize.NAME_TOKEN+ " attribute found for "+parent.getClass());
				}
				//if name -> (plug & return null); if no name (return me).
				//For primitives, call addAttr() twice - initialAttr is first one.
				//26May - now initialAttr is the initial value (whether called once
				//or twice.
				//try {
					currentGetField.putObj(name, me);	//prim/obj?
					Field f = getAllField(parent.getClass(), name);
					setFinal(f, parent, me);
					//f.set(parent, me);
				//} catch (NoSuchFieldException e) {
				//}
					//return null;	//a signal: already been dealt with...
				//gets here if: has parent, and has attr.value, and has name
 				return new NamedObject(name, me, named);
			} else { //!named -> (internal: parent, but no name)
 				return new NamedObject(name, me, named);
			}
		} else {	//parent==null -> (root, array, hash, vect)
 			return new NamedObject(name, me, named);
		}

	}


//Why is the following code still here?
	/**--------------------------------------------------------------------------
		* getArrayType: "ArrayOf-int" to "LI"
		* ------------
		* convert Array name string from:
		*		[*		//any number of '[' (including 0)
		*		(B|C|D|F|I|J|S|Z) |		//one of these represeting a primitive type
		*   (L[^;]*;)		//or the name of the class, in between 'L' and ';'
		**/
/*
	private static String getArrayType(Class clazz) {
		String a = "";	//will hold result
		String name = clazz.getName();	//NOTE: would need to escapeDollar
		if (DEBUG) System.err.println("Array type = \""+name+"\"");
		int i;
		for (i=0; name.charAt(i)=='['; i++)
			a += ARRAY_TOKEN;
		switch (name.charAt(i++)) {
			case 'B': a += "byte";
				break;
			case 'C': a += "char";
				break;
			case 'D': a += "double";
				break;
			case 'F': a += "float";
				break;
			case 'I': a += "int";
				break;
			case 'J': a += "long";
				break;
			case 'S': a += "short";
				break;
			case 'Z': a += "boolean";
				break;
			case 'L': a += name.substring(i, name.lastIndexOf(';'));
						//could also be name.length()-1
			break;
		}
		// NB: in a multi-dimensional array, Class.getComponentType() returns
		// the next one further in (which may also be an array).
		return a;
	}
*/



	/**--------------------------------------------------------------------------t
		* addAttr (map to primitives and String)
		* -------
		* Handles: byte, char, short, int, long, float, double and String.
		* As fields, arrays and custom data.
		* NOTE: String can never appear as custom data (can only be written as an
		* object).
		*
		* Consider: BIZARRE: Could do something odd for arrays: take the object
		* only as indicating the exact type of the array, and in here, when read
		* the length, actually create it.  And yet it makes sense in a way.
		*
		* or, return a proper "attrMsg", which includes the length as info.  This
		* is wha should be returned, anyway - not this hacked use of Attr.
		**/
	private ParserXML.Attr addAttr(Object o, boolean isArray) throws
			IOException,
			//NoSuchFieldException,
			ClassNotFoundException,
			ParserXML.ExceptionXML {
		ParserXML.Attr[] addAttrTmp = addName();	//split it up
		return addPrim(addAttrTmp, o, isArray);
	}

	/**--------------------------------------------------------------------------t
		* addName
		* -------
		* Functionality factored out of addAttr, for the sake of super protocol.
		*
		**/
	private ParserXML.Attr[] addName() throws
			IOException,
			//NoSuchFieldException,
			ClassNotFoundException,
			ParserXML.ExceptionXML {
		ParserXML.Attr attr;	//it *will* be set
		ParserXML.Attr name = null; //it might not be set
		if ( !(attr=p.readAttr()).isEnd ) {
			if (DEBUG) System.err.println(attr);
			if (attr.name.equals(XMLSerialize.NAME_TOKEN)) { //if name of this class
				if (name!=null)
					throw new ParserXML.ExceptionXML("just one name field");
				name = attr;
				//if (name.value.equals(XMLSerialize.SUPER_TOKEN))	//not field, an action
				//	meClassFrame = Class.forName(ParseUtilities.descapeDollar(elementName));
				//continue;	//next attr
			}
		}
		return new ParserXML.Attr[] {attr, name};
	}

	private ParserXML.Attr addPrim(ParserXML.Attr[] addAttrTmp, Object o, boolean isArray) throws
			IOException,
			//NoSuchFieldException,
			ClassNotFoundException,
			ParserXML.ExceptionXML {
		//Class meClassFrame = o.getClass();		//if not superVersion
		int length = -1;	//to flag if not set
		if (ADDATTR_NULL_DEBUG)
			System.err.println("addAttrTmp = "+addAttrTmp);
		ParserXML.Attr attr = addAttrTmp[0];	//now passed in above

		ParserXML.Attr name = addAttrTmp[1];	//now passed in above
	 if (!attr.isEnd) {
		//while (!(attr=p.readAttr()).isEnd)  //cont from addName
		do {
			if (name==attr)	continue;	//only go on if not yet processed
				//ignore all fields of array, but for the name
			if (isArray) {	//length or primitive attributes
				if (attr.name.equals(XMLSerialize.LENGTH_TOKEN)) {	//if length
					if (INTERNAL_DEBUG) {
						System.err.println("ARRAY LENGTH: " +attr);
					}
					length = Integer.parseInt(attr.value);
					break;		//break off part way through (will this work?)
				}
				if (INTERNAL_DEBUG) {
					System.err.println("ARRAY PRIMITIVE ATTRIBUTES: " +attr);
				}
				//else, continues on, reading primitive values, and sticking them in.
				//assumed: .startsWith(XMLSerialize.ARRAY_PRIMITIVE_INDEX_TOKEN)
				String indexStr = attr.name.substring(XMLSerialize.ARRAY_PRIMITIVE_INDEX_TOKEN.length());
				int index = Integer.parseInt(indexStr);	//we know it is an int
				if (DEBUG) System.err.println(attr);
				Object value = getValue(o.getClass().getComponentType(), attr);
					//This also puts alias for String, because we can infer the type from
					//the field.  Not so with custom data.  I need to write this up.  The
					//immediate result is that we serialize internal Strings like objects
					//(in tags), like primitives (in attributes).
				Array.set(o, index, value);
			}
			else {	//if not array...
				if (attr.name.endsWith(XMLSerialize.ALIAS_STRING_TOKEN)) {//string alias
					if (INTERNAL_DEBUG) {
						System.err.println("STRING ALIAS: " +attr);
					}
					//(1). extract the fieldName
					String fieldName = attr.name.substring(0, attr.name.indexOf(XMLSerialize.ALIAS_STRING_TOKEN));
					//(2). get the value & use it to lookup the object
					Object strObj = aliasHash.get(attr.value);	//get the object (String)
					//(4). assign that object to the field
					//try {
						currentGetField.putObj(fieldName, strObj);	//prim/obj?
						Field f = getAllField(o.getClass(), fieldName);
						//Field f = getAllField(meClassFrame, fieldName);
						Class fc = f.getType();			//type specified by java, not by xml
						if (DEBUG) System.err.println("Made Field \""+fc.getName()+"\"");
						setFinal(f, o, strObj);
							//just pass discard into setFinal?  That should catch everything!
					//} catch (NoSuchFieldException e) {
					//}
					//f.set(o, strObj);	//note: referring to already XML decoded object
				}
				else {	//if not string alias...
					if (attr.name.startsWith(XMLSerialize.INTERNAL_PRIMITIVE_TOKEN)) {
						if (INTERNAL_DEBUG) {
							System.err.println("INTERNAL: " +attr);
						}
						String primString = attr.name.substring(XMLSerialize.INTERNAL_PRIMITIVE_TOKEN.length());	//omit prefix (typically '_' or "_." or "someclass.")
						if (DEBUG) System.err.println(attr);
							//non-XML: we rely on the ordering of attributes
							//need (perhaps) to detect non-ordering, and at that time
							//start allowing for it....
							//or, we could detect the size, and add in nulls to match it.
							//And signal an error later.
							//NB: I think that Strings will always be done as tags, even
							//though attributes might be neater.
						primStore.add(attr.value);	//keep it in the universal form of "string"
						//IF we do readUTF() as a primitive, then we need to test the
						//type of it here, and put it into 
						//PROBLEM!!! We cannot tell the type of attributes!!!  Thus, we
						//cannot handle Strings differently!  We *could* add type
						//information, in the same hack that we use for alias strings; but
						//this is extra work, and is ugly.  We simply don't know until
						//the readObject() code is run, and the present buffer method
						//means that we cannot tell at all, unless we kept track of the
						//exact order of all custom data written... but this is lost 
						//in communication, because of the split between attr and tags...
						//Thus, we must change the format to do this, either utterly (so
						//that the order tells us), or slightly, to include type information
						//for Strings that are custom data.  Need to write a little white
						//paper on this!
							//Question: how are internal Strings stored?
							//It doesn't matter, except that they will be read with
							//readObject, and so should not be on the primStore...
							//There is also an alias issue...
					}
					else {	//if an ordinary field (not an array at all)
						if (INTERNAL_DEBUG) {
							System.err.println("ordinary: "+attr);
						}
						//try {
							if (IBM_DEBUG) System.err.println("IBM: currentGetField = "+currentGetField);
							if (IBM_DEBUG) System.err.println("IBM: attr = "+attr);
							currentGetField.putObj(attr.name, attr.value);	//prim/obj?
							if (IBM_DEBUG) System.err.println("IBM: o is "+
								(o==null?"null":"not null") );
							//Class wolfClazz = o.getClass();
							//String wolfName = attr.name;	//sep lines, for debug
							//Field f = getAllField(wolfClazz, wolfName);
							Field f = getAllField(o.getClass(), attr.name);	//subclass Field?
							//Field f = getAllField(meClassFrame, attr.name);
							Class fc = f.getType();			//type specified by java, not by xml
							if (DEBUG) System.err.println("Made Field \""+fc.getName()+"\"");
							setFinal(f, o, getValue(fc, attr));
							//f.set(o, getValue(fc, attr));
						//} catch (NoSuchFieldException e) {
						//}
					}
				}
			}
		} while (!(attr=p.readAttr()).isEnd);
   }
			//combine info from the name and final attr returned.  Awkward.
		attr.length = length;
		if (name==null) {
				//attr.value = null;	//some paths assume the name is here... ugly
			attr.nameMissing = true;	
			return attr;	//simple.  No name to return
		}
			//else, combine info from both (name and last attr)
		name.isEnd = attr.isEnd;
		name.emptyTag = attr.emptyTag;
		name.length = attr.length;
		return name;
	}


	/**--------------------------------------------------------------------------
		* getValue
		* ------------------
		* In: the class of the primitive; the attr (name-value pair)
		* Out: * object representing the primitive
		* Class may come from the "TYPE" field, or from array component type.
		* Used by:
		*		deserializeWrapper
		*		addAttr - note: an intermediate wrapper object is created redundantly
		*
		* It may be more efficient to not reuse this code.
		* when used for an attribute, it creates an Object needlessly;
		* however, such an object is required for Hashtable
		**/
	private Object getValue(Class fc, ParserXML.Attr attr) throws IOException {
		if (DEBUG) System.err.println("getting value of " +fc);
		if (fc.isPrimitive()) {     //PRIMITIVE
			if (fc==int.class) return new Integer(attr.value);	//most likely first
			else if (fc==double.class)
				return new Double(ParseUtilities.parseDouble(attr.value)); //second pr
			else if (fc==boolean.class) {
					return new Boolean(ParseUtilities.parseBoolean(attr.value));
			}
			else if (fc==byte.class) return new Byte(attr.value);
			else if (fc==char.class) {
				String a = ParseUtilities.decodeXML(attr.value);	//many chars may go to one
				if (a.length()!=1) throw new IllegalArgumentException("Character data must be exactly one character long; instead we got: '" +a+ "'");
				return new Character(a.charAt(0));
			}
			else if (fc==short.class) return new Short(attr.value);
			//else if (fc==int.class) return new Integer(attr.value));
			else if (fc==long.class) return new Long(attr.value);
			else if (fc==float.class)
				return new Float(ParseUtilities.parseFloat(attr.value));
			else throw new ParserXML.ExceptionXML(
				"Unimplemented primitive: \"" +fc.getName()+ "\"");
		}
		else {                      //CLASS
			if (fc==String.class) {
				String s = ParseUtilities.decodeXML(attr.value);
				putAlias(s);	//String is an object - not called for attr string
				return s;
			}
			else throw new ParserXML.ExceptionXML(
				"'"+attr.name+"' is not primitive or String");
		}
	}




	/**--------------------------------------------------------------------------
		* addVectorElements
		* -----------------
		* Note: this should now be possible automatically, since we can read
		* private fields (and any internal serialization)
		* Note: Vector object already created when we get here.
		**/
  private void addVectorElements(Vector v) throws
			ClassNotFoundException,
			//NoSuchFieldException,
			IOException {
		ParserXML.Tag tag;
		while (true) {
			Object obj;
			tag = p.readTag();
			if (tag==null) throw new EOFException("Premature EOF");
			if (!tag.start) return;	//done!  caller will check if a correct endtag
			if (tag.name.equals("java.lang.String")	//bad: doubles up
					|| tag.name.equals("java.lang.Integer")
					|| tag.name.equals("java.lang.Double")
					|| tag.name.equals("java.lang.Boolean")
					|| tag.name.equals("java.lang.Character")
					|| tag.name.equals("java.lang.Byte")
					|| tag.name.equals("java.lang.Short")
					|| tag.name.equals("java.lang.Long")
					|| tag.name.equals("java.lang.Float")
				)
				obj = deserializeWrapper(tag, null).object;
			else
				//obj = createObject(tag.name, null);	//null=orphan object
				obj = createObject(tag, null);	//null=orphan object //30Apr2001
			if (DEBUG) System.err.println("Adding {"+obj+"} to Vector");
			v.add(obj);
		}
		//if (DEBUG) System.err.println("Vector: \n" +v);
	}

	/**--------------------------------------------------------------------------
		* addHashtableElements
		* --------------------
		* An container object.
		* Handles just tags (no attrs)
		**/
	private void addHashtableElements(Hashtable h) throws
			ClassNotFoundException,
			//NoSuchFieldException,
			IOException {
		ParserXML.Tag tag;
		while (true) {
			Object[] objs = {null, null};
			for (int i=0; i<objs.length; i++) {		//{key, value}
				tag = p.readTag();
				if (tag==null) throw new EOFException("Premature EOF");
				if (!tag.start)
					if (i==0) return; //end of hashtable
					else throw new IOException("incomplete Hashtable entry");
				if (tag.name.equals("java.lang.String")	//bad: doubles up
						|| tag.name.equals("java.lang.Integer")
						|| tag.name.equals("java.lang.Double")
						|| tag.name.equals("java.lang.Boolean")
						|| tag.name.equals("java.lang.Character")
						|| tag.name.equals("java.lang.Byte")
						|| tag.name.equals("java.lang.Short")
						|| tag.name.equals("java.lang.Long")
						|| tag.name.equals("java.lang.Float")
					)
					objs[i] = deserializeWrapper(tag, null).object;
				else
					//objs[i] = createObject(tag.name, null);	//null=orphan object
					objs[i] = createObject(tag, null);	//null=orphan object//30Apr2001
			}
			if (DEBUG) System.err.println("Putting {"+objs[0]+", "+objs[1]+"}");
			h.put(objs[0], objs[1]);
		}
		//if (DEBUG) System.err.println("Hashtable: \n"+h);
	}

	/**--------------------------------------------------------------------------
		* deserializeClass
		* ----------------
		* Handles the attrs.
		* Need to refactor the test for wrapper type
		* Reads the whole thing in, including <x></x> or <x/>
		**/
	private NamedObject deserializeClass(ParserXML.Attr[] addAttrTmp, ParserXML.Tag startTag, Object cheatParent) throws
			ClassNotFoundException,
			//NoSuchFieldException,
			IOException {
		ParserXML.Attr attr = addAttrTmp[0];
		ParserXML.Attr end, name = null;
		ParserXML.Tag tag;
		//attr = p.readAttr();	//p is the ParserXML.
		if (DEBUG) System.err.println("deserializeClass: "+ attr);
		if (attr.name.equals(XMLSerialize.NAME_TOKEN)) {
			//System.err.println("Got a name in a wrapper: "+startTag.name+" "+attr.value+";");
			name = attr;
			attr = p.readAttr();
		}
		if (!attr.name.equals(XMLSerialize.CLASSNAME_TOKEN)) 
			throw new ParserXML.ExceptionXML(XMLSerialize.CLASSNAME_TOKEN+ " expected");
		if (!(end = p.readAttr()).isEnd)
			throw new ParserXML.ExceptionXML("\">\" or \"/>\" expected");
		if (!end.emptyTag) {
			tag = p.readTag();
			if (!tag.name.equals(startTag.name) || tag.start)
				throw new ParserXML.ExceptionXML("</"+tag.name+"> expected");
		}
		//IDEA: could make up the expected tag, and use Tag.equals() to compare
		if (DEBUG) System.err.println("String read in: " + attr.value);

		String className = ParseUtilities.descapeDollar(attr.value);
		Class clazz = null;
		try {
			clazz =  Class.forName(className);	//won't find primitives
				//javadocs: "If name denotes a primitive type or void, an attempt will"
				//"be made to locate a user-defined class in the unnamed package whose"
				//"name is name. Therefore, this method cannot be used to obtain any"
				//"of the Class objects representing primitive types or void."
			//bug scratch patch submitted by: matthew sandoz, Nov 2, 2001

		} catch (ClassNotFoundException e) {
			if (className.equals("int"))          clazz = int.class;
			else if (className.equals("double"))  clazz = double.class;
			else if (className.equals("boolean")) clazz = boolean.class;
			else if (className.equals("byte"))    clazz = byte.class;
			else if (className.equals("char"))    clazz = char.class;
			else if (className.equals("short"))   clazz = short.class;
			else if (className.equals("long"))    clazz = long.class;
			else if (className.equals("float"))   clazz = float.class;
			else if (className.equals("void"))    clazz = void.class;
			else
				throw new ClassNotFoundException("class '"+className+"' not found");
		}

		putAlias(clazz);	//String is an object
		if (name!=null)	{ //can only happen for Strings... (ugly hack!)
			//try {
				currentGetField.putObj(name.value, clazz);	//prim/obj?
				Field f = getAllField(cheatParent.getClass(), name.value);
				setFinal(f, cheatParent, clazz);
				//f.set(cheatParent, clazz);	//Custom "Class" object code...
			//} catch (NoSuchFieldException e) {
					//evolution policy is to fail silently.
			//}
 			return new NamedObject(name.value, clazz, true); //name!=null -> true
		}
 		return new NamedObject(null, clazz, false); //name!=null -> false
	}


	/**--------------------------------------------------------------------------
		* deserializeBinary
		* ----------------
		* Handles the attrs.
		* Need to refactor the test for wrapper type
		* Reads the whole thing in, including <x></x> or <x/>
		**/
	private NamedObject deserializeBinary(ParserXML.Attr[] addAttrTmp, ParserXML.Tag startTag, Object cheatParent) throws
			ClassNotFoundException,
			//NoSuchFieldException,
			IOException {
		ParserXML.Attr attr = addAttrTmp[0];
		ParserXML.Attr end, name = null;
		ParserXML.Tag tag;
		//attr = p.readAttr();	//p is the ParserXML.
		if (DEBUG) System.err.println("deserializeBinary: "+ attr);
		if (attr.name.equals(XMLSerialize.NAME_TOKEN)) {
			//System.err.println("Got a name in a wrapper: "+startTag.name+" "+attr.value+";");
			name = attr;
			attr = p.readAttr();
		}
		if (!attr.name.equals(XMLSerialize.VALUE_TOKEN)) 
			throw new ParserXML.ExceptionXML(XMLSerialize.VALUE_TOKEN+ " expected");
		if (!(end = p.readAttr()).isEnd)
			throw new ParserXML.ExceptionXML("\">\" or \"/>\" expected");
		if (!end.emptyTag) {
			tag = p.readTag();
			if (!tag.name.equals(startTag.name) || tag.start)
				throw new ParserXML.ExceptionXML("</"+tag.name+"> expected");
		}
		//IDEA: could make up the expected tag, and use Tag.equals() to compare
		if (DEBUG) System.err.println("Binary data read in: " + attr.value);

		byte[] binaryData = ParseUtilities.decodeHex(attr.value);
		putAlias(binaryData);	//String is an object
		if (name!=null)	{ //can only happen for Strings... (ugly hack!)
			//try {
				currentGetField.putObj(name.value, binaryData);	//prim/obj?
				Field f = getAllField(cheatParent.getClass(), name.value);
				setFinal(f, cheatParent, binaryData);
				//f.set(cheatParent, binaryData);	//
			//} catch (NoSuchFieldException e) {
					//evolution policy is to fail silently.
			//}
 			return new NamedObject(name.value, binaryData, true); //name!=null -> true
		}
 		return new NamedObject(null, binaryData, false); //name!=null -> false
	}


	/**--------------------------------------------------------------------------
		* deserializeWrapper
		* ------------------
		* Handles the attrs.
		* Need to refactor the test for wrapper type
		* Reads the whole thing in, including <x></x> or <x/>
		**/
	private NamedObject deserializeWrapper(ParserXML.Tag startTag, Object cheatParent) throws
			ClassNotFoundException,
			//NoSuchFieldException,
			IOException {
			return deserializeWrapper(addName(), startTag, cheatParent);
	}
	private NamedObject deserializeWrapper(ParserXML.Attr[] addAttrTmp, ParserXML.Tag startTag, Object cheatParent) throws
			ClassNotFoundException,
			//NoSuchFieldException,
			IOException {
		ParserXML.Attr attr = addAttrTmp[0];
		ParserXML.Attr end, name = null;
		ParserXML.Tag tag;
		//attr = p.readAttr();	//p is the ParserXML.
		if (DEBUG) System.err.println("deserializeWrapper: "+ attr);
		if (attr.name.equals(XMLSerialize.NAME_TOKEN)) {
			//System.err.println("Got a name in a wrapper: "+startTag.name+" "+attr.value+";");
			name = attr;
			attr = p.readAttr();
		}
		if (!attr.name.equals(XMLSerialize.VALUE_TOKEN)) 
			throw new ParserXML.ExceptionXML(XMLSerialize.VALUE_TOKEN+ " expected");
		if (!(end = p.readAttr()).isEnd)
			throw new ParserXML.ExceptionXML("\">\" or \"/>\" expected");
		if (!end.emptyTag) {
			tag = p.readTag();
			if (!tag.name.equals(startTag.name) || tag.start)
				throw new ParserXML.ExceptionXML("</"+tag.name+"> expected");
		}
		//IDEA: could make up the expected tag, and use Tag.equals() to compare
		if (DEBUG) System.err.println("String read in: " + attr.value);
		if (startTag.name.equals("java.lang.String")) {	
			String s = ParseUtilities.decodeXML((String)attr.value);	//not attr.name! Don't forget to decode!
			putAlias(s);	//String is an object
			if (name!=null)	{ //can only happen for Strings... (ugly hack!)
				//try {
					currentGetField.putObj(name.value, s);	//prim/obj?
					Field f = getAllField(cheatParent.getClass(), name.value);
					setFinal(f, cheatParent, s);
					//f.set(cheatParent, s);
				//} catch (NoSuchFieldException e) {
				//}
 				return new NamedObject(name.value, s, true); //name!=null -> true
			}
 			return new NamedObject(null, s, false); //name!=null -> false
		}
		Class c = Class.forName(ParseUtilities.descapeDollar(startTag.name));
		try {
		//NOTE: We had best make this work directly, rather than via reflection,
		//because then we can use static type checking rather than messy
		//exceptions (which we'd rather not have to deal with!)
		//Convert from Integer.class (Wrapper type) to int.class (primitive type)
			Field f = c.getDeclaredField("TYPE");	//will work for wrappers
				//Field f = getAllField(c, "TYPE");	//a cheap way to get the type
			Object o = getValue((Class)f.get(c), attr);
			putAlias(o);	//NOTE: *not* called if a String, so not inc twice
 			return new NamedObject(null, o, false); //can never be named
		} catch (NoSuchFieldException e) {
			//Special case: Wrappers should have a field named "TYPE".
			System.err.println("'" +c+ "' is not a wrapper class: "+e);
		} catch (IllegalAccessException e) {
			System.err.println("'" +c+ "' is not a wrapper class: "+e);
		}
		throw new IOException("Shouldn't get here");
	}



	/**--------------------------------------------------------------------------
		* Test class
		* ----------
		* Predates ObjIn and ObjOut!
		**/
	public static class Test {
		public static void main(String args[]) throws Exception {

			XMLDeserialize d = new XMLDeserialize();	//defaults to System.in
			XMLSerialize s = new XMLSerialize();	//defaults to System.out
			Object o;
			while ( (o = d.deserialize()) != null) {	//can loop for streamed XML
			/*
				System.err.println("Printing created object:");
				System.err.println();
				System.err.println(o);
				System.err.println();
				System.err.println("(Re)serializing created object:");
				System.err.println();
			*/
				s.serialize(o);
			}
		}
	}
}
