
/*
 * Copyright (C) 2000-2001 Technical University of Ilmenau, Germany
 *                         Department of Microelectronic Circuits and Systems
 *
 * You may only use this file according to the terms of the license agreement included
 * within the file LICENSE which is distributed together with the software package.
 *
 * CVS Info:
 * $Author: rummler $
 * $Date: 2001/11/19 13:32:05 $
 * $Revision: 1.5 $
 * $State: Exp $
 * $RCSfile: ListInitialization.java,v $
 *
 */

package mss.ea.ex.mutation;

import mss.ea.chr.*;
import mss.ea.ini.*;

/** Simple initialization operator.
 * This operator creates integer arrays.
 * 
 * @since 0.2
 * @author Andreas Rummler
 * @version $Revision: 1.5 $ $Date: 2001/11/19 13:32:05 $
 */
public class ListInitialization extends Initialization {
	
	/** Size of the integer array.
	 */
	protected int arraySize = 1;
	
	/** Constructor.
	 * 
	 * @param indNumber number of individuals to be generated
	 * @param arraySize size of the integer array
	 */
	public ListInitialization( int indNumber, int arraySize ) {
		super( indNumber );
		this.arraySize = arraySize;
	}
	
	/** Creates a chromosome set containing one single ArrayChromosome.
	 * 
	 * @return newly created chromosome set
	 */
	public ChromosomeSet createSet() {
		// create a new set
		ChromosomeSet set = new ChromosomeSet();
		int[] array = new int[arraySize];
		// create a valid chromosome set
		for ( int i = 0; i < arraySize; i++ ) {
			array[i] = i;
		}
		ArrayChromosome ac = new ArrayChromosome( array );
		set.add( ac );
		return set;
	}
	
}
