
/*
 * Copyright (C) 2000-2001 Technical University of Ilmenau, Germany
 *                         Department of Microelectronic Circuits and Systems
 *
 * You may only use this file according to the terms of the license agreement included
 * within the file LICENSE which is distributed together with the software package.
 *
 */

package mss.ea.ex.sphere;

import java.util.*;

import mss.ea.al.*;
import mss.ea.core.*;
import mss.ea.eval.*;
import mss.ea.ins.*;
import mss.ea.mut.*;
import mss.ea.rec.*;
import mss.ea.stat.*;
import mss.ea.trm.*;

public class Sphere extends AbstractAlgorithm {
	
	protected int dimensions = 3;
	
	// overload setup
	protected void setup() {
		
		// create initialization (source)
		SphereInit init = new SphereInit( 10, dimensions );
		// operator for calculation of the sphere function
		SphereCalc eval = new SphereCalc();
		SphereCalc eval2 = new SphereCalc();
		// generic population
		SortedPopulation gate = new SortedPopulation( new DefaultIndividualComparator( new ReverseScoreComparator() ) );
		gate.setFilter( new EliteReinsertion2( (float) 0.5 ) );
		// create array of chromsome mutation operators
		ChromosomeMutation[] cma = new ChromosomeMutation[dimensions];
		for ( int i = 0; i < dimensions; i++ ) {
			cma[i] = new DoubleRelativeRangeMutation( (float) 1.0, (float) 0.2 );
		}
		// duplication mutation operator
		IndividualDuplication dup = new IndividualDuplication( 1 );
		// create mutation operator
		IndividualMutation mut = new IndividualMutation( cma );
		// create sink
		GenericSink snk = new GenericSink();
		
		// define termination conditions
		// maximal number of generations : 500
		Condition genCond = new CounterCondition( gate.getCounter(), 500 );
		ConditionalRouter cond = new ConditionalRouter( genCond );
		
		DefaultIndividualCollector coll = new DefaultIndividualCollector();
		
		// add a notifier
//		EventNotifier not = new EventNotifier( gate );
//		DefaultDataTracker ddt = new DefaultDataTracker( "ddt.log", true );
//		PopulationSizeCollector psize = new PopulationSizeCollector( ddt, (AbstractPopulation) gate.getGate() );
//		not.addListener( psize );
//		not.addListener( ddt );
		
		connect( init, eval );
		connect( eval, coll, 0 );
		connect( coll, gate );
		connect( gate, cond );
//		connect( not, cond );
		connect( cond, 0, dup );
		connect( cond, 1, snk );
		connect( dup, mut );
		connect( mut, eval2 );
		connect( eval2, coll, 1 );
//		connect( ins, gate );
		
	}
	
	public static void main( String[] args ) throws Exception {
		Sphere s = new Sphere();
		s.start( s, args );
	}
	
}
