import java.io.*;

/**
	* This class demonstrates PutField, GetField and serialPersistentFields
	* as a way to allow the XML represention of a class to differ from the
	* actual fieldnames in the class.	
	* For example, your internal implementation could evolve utterly, while
	* maintaining back-compatility with the names used to serialize the class
	* originally - this is "Schema Evolution" (since a class is a type is a
	* schema).
	* These mechanisms come from the Serialization API, so your code can be
	* dynamically swapped between binary (performance, cryptic) and XML
	* (slower, longer, but readable and interoperable).

	* USAGE with JSX:
	* ---------------
	*   java Evolve > t.xml               //create XML output
	*   java JSX.ObjIn < t.xml > tt.xml   //see if it recreates same...
	* t.xml and tt.xml should be identical same...
	* 
	* USAGE with JOS:  (Java's Own Serialization)
	* ---------------
	*   java Evolve JOS > t.ser            //create binary output
	*   java Evolve$JOS < t.ser > tt.ser   //see if it recreates same...
	* t.ser and tt.ser should be identical...
	*
	* This class is very first stab, but we are "releasing early" by he open
	* source mantra, to make sure we are on the right track, to expose bugs and
	* any deep design problems early, and to get the satisfaction of getting
	* something working out the door.
	*
	* Limitations: Nested classes that use the PutField/GetField API will
	* be confused between levels of nesting in particular circumstances.  This
	* will be easy to fix in the next release, by simply stacking.
	**/


public class Evolve implements Serializable {

	public static void main(String args[]) throws Exception {
		ObjectOutputStream out;
		if (args.length>0) out = new ObjectOutputStream(System.out);
		else out = new JSX.ObjOut(System.out);

		Object o = new Evolve();
		out.writeObject(o);
		out.flush();	//JOS needs this

	}

	static class JOS {	//Java's Own Serialization
		public static void main(String args[]) throws Exception {
			ObjectInputStream in = new ObjectInputStream(System.in);
			Object obj = in.readObject();
			ObjectOutputStream out = new ObjectOutputStream(System.out);
			out.writeObject(obj);	//so we can see it

			out = new JSX.ObjOut(System.err);
			out.writeObject(obj);	//so we can see it
		}
	}

	static final long serialVersionUID = 30593813711490L;
	static final String a = "a";
	static final String b = "b";
	static final String c = "c";
	String xChanged = "hello";
	//String x = "hello";
	int y = 5;
	Link z = Link.root;
/*
	private static final ObjectStreamField[] serialPersistentFields = { 
 		new ObjectStreamField(a, String.class),
 		new ObjectStreamField(b, Integer.TYPE),
 		new ObjectStreamField(c, Link.class),
	}; 
	private void writeObject(ObjectOutputStream s) throws IOException {
		ObjectOutputStream.PutField fields = s.putFields();
		fields.put(a, x);
		fields.put(b, y);
		fields.put(c, z);
		s.writeFields(); //alt: fields.write(s);
	}
	private void readObject(ObjectInputStream s) throws IOException {
		ObjectInputStream.GetField fields = null;
		try {
			fields = s.readFields();
		} catch (Exception ClassNotFoundException) {
 			throw new IOException("ClassNotFoundException wrapped");
		}
		x = (String) fields.get(a, "we got nothing!");
		y = fields.get(b, -100);
		z = (Link) fields.get(c, new Link());
	}
*/
}



class Link implements Serializable {
	static Link root;
	static {
		root = new Link();
		root.next = new Link();
		root.next.next = new Link();
		root.next.next.next = root;
	}
	Link next;

	static final long serialVersionUID = 3059381371L;
	static final String a = "NEXT";
	static final String b = "next";	//we (hope) we cope with synonyms...
	private static final ObjectStreamField[] serialPersistentFields = { 
 		new ObjectStreamField(a, Link.class),
	}; 
	private void writeObject(ObjectOutputStream s) throws IOException {
		ObjectOutputStream.PutField fields = s.putFields();
		fields.put(a, next);
		s.writeFields(); //alt: fields.write(s);
	}
	private void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
		ObjectInputStream.GetField fields = null;
		//try {
			fields = s.readFields();
		//} catch (Exception ClassNotFoundException) {
 		//	throw new IOException("ClassNotFoundException wrapped");
		//}
		if (!fields.defaulted(a))
			next = (Link) fields.get(a, null);	//default to null
		else if (!fields.defaulted(b))
			next = (Link) fields.get(b, null);	//default to null
		//nested GetFields are not implemented correctly at present (need to stack)
	}
}
